#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


import os
import os.path as op
import json
import traceback
from datetime import datetime
from elasticsearch.exceptions import NotFoundError
from aliyun.log import LogClient
from aliyun.log.putlogsrequest import PutLogsRequest
from aliyun.log.es_migration.doc_logitem_converter import DocLogItemConverter


class Checkpoint(object):

    processing = 'processing'
    interrupted = 'interrupted'
    finished = 'finished'
    dropped = 'dropped'
    failed = 'failed'

    def __init__(
            self, ckpt_path, task_id, es_index, es_shard,
            logstore, time_reference, logger,
    ):
        self._ckpt_file = op.join(ckpt_path, 'task_{}.json'.format(task_id))
        self._logger = logger
        self._content = {
            'status': self.processing,
            'task': {
                'es_index': es_index,
                'es_shard': es_shard,
                'logstore': logstore,
                'time_reference': time_reference,
            },
            'update_time': None,
            'progress': 0,
            'checkpoint': {
                '_id': None,
                'scroll_id': None,
            },
        }
        self._load()

    @property
    def id(self):
        return self._content['checkpoint']['_id']

    @property
    def scroll_id(self):
        return self._content['checkpoint']['scroll_id']

    @property
    def status(self):
        return self._content['status']

    @property
    def content(self):
        return self._content

    def update(self, status=processing, count=0, _id=None, scroll_id=None):
        now = datetime.now()
        self._content['update_time'] = now.isoformat(timespec='seconds')
        self._content['status'] = status
        if _id:
            self._content['checkpoint']['_id'] = _id
        self._content['progress'] += count
        if scroll_id:
            self._content['checkpoint']['scroll_id'] = scroll_id
        ckpt_bak = self._ckpt_file + '.bak'
        try:
            os.rename(self._ckpt_file, ckpt_bak)
        except FileNotFoundError:
            pass
        with open(self._ckpt_file, 'w') as fp:
            fp.write(json.dumps(self._content, indent=2))
        try:
            os.remove(ckpt_bak)
        except FileNotFoundError:
            pass

    def _load(self):
        try:
            with open(self._ckpt_file) as f:
                cont = f.read().strip()
        except FileNotFoundError:
            cont = ''
        if len(cont) > 0:
            try:
                cont = json.loads(cont)
            except json.JSONDecodeError:
                msg = 'Invalid checkpoint file content'
                extra = {'cache': cont}
                self._logger.error(msg, extra)
                raise Exception(msg)

            if cont['task'] != self._content['task']:
                msg = 'Specified task not matches with cache'
                extra = {'task': self._content['task'], 'cache': cont['task']}
                self._logger.error(msg, extra)
                raise Exception(msg)

            self._content.update(cont)


class MigrationTask(object):

    def __init__(
            self, _id, es_client, es_index, es_shard,
            logstore, ckpt_path, batch_size, logger,
            es_version, es_query=None, es_scroll="1d", time_reference=None,
    ):
        self._id = _id
        self._es_client = es_client
        self._es_index = es_index
        self._es_shard = es_shard
        self._es_query = es_query or {}
        self._es_scroll = es_scroll
        self._logstore = logstore
        self._time_reference = time_reference
        self._batch_size = batch_size
        self._logger = logger
        self._ckpt = Checkpoint(
            ckpt_path,
            self._id,
            self._es_index,
            self._es_shard,
            self._logstore.name,
            self._time_reference,
            self._logger,
        )
        self._es_params = {'preference': '_shards:{:d}'.format(self._es_shard)}
        self._es_scroll_id = self._ckpt.scroll_id
        self._status = self._ckpt.status
        self._cnt, self._last = 0, None

    def run(self, shutdown_flag):
        # already finished
        if self._ckpt.status == Checkpoint.finished:
            self._logger.info('Already finished. Ignore it.')
            return Checkpoint.finished
        self._logger.info('Migration task starts', extra=self._ckpt.content)
        try:
            self._run(shutdown_flag)
        except NotFoundError:
            self._ckpt.update(status=Checkpoint.dropped)
            self._logger.info(
                'ES index dropped',
                extra={'traceback': traceback.format_exc()},
            )
        except KeyboardInterrupt:
            self._logger.info('Migration interrupted')
            self._status = Checkpoint.interrupted
        except BaseException:
            self._logger.error(
                'Exception',
                extra={'traceback': traceback.format_exc()},
            )
            self._status = Checkpoint.failed
        finally:
            self._ckpt.update(
                status=self._status,
                count=self._cnt,
                _id=None if self._last is None else self._last.get('_id'),
                scroll_id=self._es_scroll_id,
            )

        self._logger.info('Migration task exits', extra=self._ckpt.content)
        return self._ckpt.status

    def _run(self, shutdown_flag):
        rnd = 0
        while not op.exists(shutdown_flag):
            if self._es_scroll_id is None:
                # initial search
                resp = self._es_client.search(
                    index=self._es_index,
                    body=self._es_query,
                    scroll=self._es_scroll,
                    size=self._batch_size,
                    params=self._es_params,
                )
            else:
                try:
                    resp = self._es_client.scroll(
                        scroll_id=self._es_scroll_id,
                        scroll=self._es_scroll,
                    )
                except NotFoundError:
                    msg = f"cache is expired, which is with duration {self._es_scroll}"
                    self._logger.error(msg, exc_info=True)
                    raise Exception(msg)

            scroll_id, hits = resp.get('_scroll_id'), resp['hits']['hits']
            if len(hits) > 0:
                self._put_docs(hits)
                self._cnt += len(hits)
                self._last = hits[-1]

            # end of scroll
            if scroll_id is None or len(hits) <= 0:
                self._status = Checkpoint.finished
                self._logger.info('Migration finished')
                break

            self._es_scroll_id = scroll_id
            rnd += 1
            if rnd % 100 == 0 or rnd == 1:
                _id = self._last.get('_id')
                self._ckpt.update(count=self._cnt, _id=_id, scroll_id=scroll_id)
                self._cnt = 0
                self._logger.info('Migration progress', extra=self._ckpt.content)

    def _put_docs(self, docs):
        if len(docs) <= 0:
            return
        logitems = [
            DocLogItemConverter.to_log_item(doc, self._time_reference)
            for doc in docs
        ]
        self._logstore.put_logs(logitems)


class MigrationLogstore(object):
    def __init__(
            self,
            endpoint,
            access_id,
            access_key,
            project_name,
            logstore_name,
            topic,
            source,
    ):
        self._log_client = LogClient(
            endpoint=endpoint,
            accessKeyId=access_id,
            accessKey=access_key,
        )
        self._project_name = project_name
        self._logstore_name = logstore_name
        self._topic, self._source = topic, source

    @property
    def name(self):
        return self._logstore_name

    def put_logs(self, logitems):
        self._log_client.put_logs(
            PutLogsRequest(
                project=self._project_name,
                logstore=self._logstore_name,
                topic=self._topic,
                source=self._source,
                logitems=logitems,
            )
        )



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logresponse import LogResponse


class ListLogstoresResponse(LogResponse):
    """ The response of the ListLogstores API from log.
    
    :type resp: dict
    :param resp: ListLogstoresResponse HTTP response body
    
    :type header: dict
    :param header: ListLogstoresResponse HTTP response header
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = resp['count']
        self.total = resp['total']
        self.logstores = resp.get('logstores', [])

    def get_count(self):
        """ Get total count of logstores from the response
        
        :return: int, the number of total logstores from the response
        """
        return self.count

    def get_total(self):
        return self.total

    def get_logstores(self):
        """ Get all the logstores from the response
        
        :return: list, all logstores
        """
        return self.logstores

    def log_print(self):
        print('ListLogstoresResponse:')
        print('headers:', self.get_all_headers())
        print('count:', self.count)
        print('total:', self.total)
        print('logstores:', self.logstores)

    def merge(self, response):
        if not isinstance(response, ListLogstoresResponse):
            raise ValueError("passed response is not a ListLogstoresResponse: " + str(type(response)))

        self.count += response.get_count()
        self.total = response.get_total() # use the latest total count
        self.logstores.extend(response.get_logstores())

        # update body
        self.body = {
            'count': self.count,
            'total': self.total,
            'logstores': self.logstores
        }

        return self



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logresponse import LogResponse


class GetCursorTimeResponse(LogResponse):
    """ The response of the get_cursor_time API from log.
    
    :type header: dict
    :param header: GetCursorTimeResponse HTTP response header

    :type resp: dict
    :param resp: the HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.cursor_time = resp['cursor_time']

    def get_cursor_time(self):
        """

        :return:
        """
        return self.cursor_time

    def log_print(self):
        print('GetCursorTimeResponse')
        print('headers:', self.get_all_headers())
        print('cursor_time:', self.cursor_time)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


import os
import os.path as op
import time
import json
import uuid
import signal
import logging
import traceback
from concurrent.futures import as_completed, ProcessPoolExecutor
from elasticsearch import Elasticsearch
from aliyun.log.logclient import LogClient, LogException
from aliyun.log.util import PrefixLoggerAdapter
from .util import split_and_strip
from .migration_log import setup_logging
from .migration_task import MigrationTask, MigrationLogstore, Checkpoint
from .index_logstore_mappings import IndexLogstoreMappings
from .mapping_index_converter import MappingIndexConverter


_logger = logging.getLogger(__name__)


class MigrationConfig(object):
    """
    Configuration for migrating data from elasticsearch to aliyun log service (SLS)

    :type cache_path: string
    :param cache_path: file path to store migration cache, which used for resuming migration process when stopped. Please ensure it's clean for new migration task.

    :type hosts: string
    :param hosts: a comma-separated list of source ES nodes. e.g. "localhost:9200,other_host:9200"

    :type project_name: string
    :param project_name: specify the project_name of your log services. e.g. "your_project"

    :type indexes: string
    :param indexes: a comma-separated list of source index names. e.g. "index1,index2"

    :type query: string
    :param query: used to filter docs, so that you can specify the docs you want to migrate. e.g. '{"query": {"match": {"title": "python"}}}'

    :type logstore_index_mappings: string
    :param logstore_index_mappings: specify the mappings of log service logstore and ES index. e.g. '{"logstore1": "my_index*", "logstore2": "index1,index2"}}'

    :type pool_size: int
    :param pool_size: specify the size of migration task process pool. Default is 10 if not set.

    :type time_reference: string
    :param time_reference: specify what ES doc's field to use as log's time field. e.g. "field1"

    :type source: string
    :param source: specify the value of log's source field. e.g. "your_source"

    :type topic: string
    :param topic: specify the value of log's topic field. e.g. "your_topic"

    :type batch_size: int
    :param batch_size: max number of logs written into SLS in a batch. SLS requires that it's no bigger than 512KB in size and 1024 lines in one batch. Default is 1000 if not set.

    :type wait_time_in_secs: int
    :param wait_time_in_secs: specify the waiting time between initialize aliyun log and executing data migration task. Default is 60 if not set.

    :type auto_creation: bool
    :param auto_creation: specify whether to let the tool create logstore and index automatically for you. e.g. True

    :type retries_failed: int
    :param retries_failed: specify retrying times for failed tasks. e.g. 10

    """

    default_pool_size = 10
    default_wait_time = 60
    default_batch_size = 1000
    default_retries_failed = 10

    def __init__(
            self,
            cache_path,
            endpoint=None,
            access_key_id=None,
            access_key=None,
            project_name=None,
            hosts=None,
            indexes=None,
            query=None,
            scroll="1d",
            time_reference=None,
            logstore_index_mappings=None,
            source=None,
            topic=None,
            pool_size=None,
            batch_size=None,
            wait_time_in_secs=None,
            auto_creation=True,
            retries_failed=None,
    ):
        self.cache_path = cache_path
        self.access_key_id, self.access_key = access_key_id, access_key

        self.ckpt_path = op.join(cache_path, 'ckpt')
        os.makedirs(self.ckpt_path, exist_ok=True)
        self._config_file = op.join(self.cache_path, 'config.json')
        self._valid = True
        if self._load_cache():
            cont = {
                'endpoint': endpoint,
                'project_name': project_name,
                'hosts': hosts,
                'indexes': indexes,
                'query': query,
                'scroll': scroll,
                'logstore_index_mappings': logstore_index_mappings,
                'source': source,
                'topic': topic,
                'pool_size': pool_size,
                'batch_size': batch_size,
                'wait_time_in_secs': wait_time_in_secs,
                'auto_creation': auto_creation,
            }
            self._cont.update({k: v for k, v in cont.items() if v is not None})
        else:
            if pool_size is None:
                pool_size = self.default_pool_size
            if batch_size is None:
                batch_size = self.default_batch_size
            if wait_time_in_secs is None:
                wait_time_in_secs = self.default_wait_time
            if retries_failed is None:
                retries_failed = self.default_retries_failed
            self._cont = {
                'endpoint': endpoint,
                'project_name': project_name,
                'hosts': hosts,
                'indexes': indexes,
                'query': query,
                'scroll': scroll,
                'time_reference': time_reference,
                'logstore_index_mappings': logstore_index_mappings,
                'source': source,
                'topic': topic,
                'pool_size': pool_size,
                'batch_size': batch_size,
                'wait_time_in_secs': wait_time_in_secs,
                'auto_creation': auto_creation,
                'retries_failed': retries_failed,
            }
        self._dump_cache()

    @property
    def valid(self):
        return self._valid

    def get(self, name):
        return self._cont.get(name)

    @property
    def hosts(self):
        return split_and_strip(self.get('hosts'), sep=',')

    def _load_cache(self):
        cached = False
        try:
            with open(self._config_file) as f:
                cont = f.read()
        except FileNotFoundError:
            cont = ''

        if len(cont) > 0:
            try:
                self._cont = json.loads(cont)
                cached = True
            except json.JSONDecodeError:
                raise Exception('Invalid migration configuration cache')
        return cached

    def _dump_cache(self):
        with open(self._config_file, 'w') as f:
            f.write(json.dumps(self._cont, indent=2))


class MigrationManager(object):
    def __init__(self, config):
        self._config = config
        _uuid = str(uuid.uuid4())
        self._id = ''.join(_uuid.split('-'))
        self._es = Elasticsearch(
            hosts=self._config.hosts,
            timeout=60,
            max_retries=30,
            retry_on_timeout=True,
            verify_certs=False,
        )
        self._es_version = int(self._es.info()['version']['number'].split('.')[0])
        self._log_client = LogClient(
            endpoint=self._config.get('endpoint'),
            accessKeyId=self._config.access_key_id,
            accessKey=self._config.access_key,
        )
        setup_logging(
            self._id,
            self._config.get('endpoint'),
            self._config.get('project_name'),
            self._config.access_key_id,
            self._config.access_key,
        )
        self._logger = logging.getLogger(__name__)
        self._shutdown_flag = op.join(self._config.cache_path, 'shutdown.lock')
        print('#migration: {}'.format(self._id))

    def migrate(self):
        self._logger.info('Migration starts')
        tasks = self._discover_tasks()
        task_cnt = len(tasks)
        pool_size = min(self._config.get('pool_size'), task_cnt)
        print('#pool_size: {}'.format(pool_size))
        print('#tasks: {}'.format(task_cnt))

        self._prepare()
        with ProcessPoolExecutor(max_workers=pool_size) as pool:
            retries_failed = self._config.get('retries_failed')
            for i in range(retries_failed):
                msg = 'Start migration tasks'
                print('\n>>>> ', msg)
                _logger.info(msg, extra={'retries': i})

                futures = []
                state = {
                    'total': task_cnt,
                    Checkpoint.finished: 0,
                    Checkpoint.dropped: 0,
                    Checkpoint.failed: 0,
                }
                for task in tasks:
                    futures.append(
                        pool.submit(
                            _migration_worker,
                            self._config,
                            task,
                            self._shutdown_flag,
                            self._es_version,
                        )
                    )
                try:
                    for future in as_completed(futures):
                        res = future.result()
                        if res in state:
                            state[res] += 1
                        self._logger.info('State', extra=state)
                        print('>> state:', json.dumps(state))
                except BaseException:
                    self._logger.error(
                        'Exception',
                        extra={'traceback': traceback.format_exc()},
                    )
                    for future in futures:
                        if not future.done():
                            future.cancel()
                    list(as_completed(futures, timeout=10))

                if op.exists(self._shutdown_flag):
                    # Already interrupted
                    break

                if state[Checkpoint.finished] + state[Checkpoint.dropped] >= task_cnt:
                    print('All migration tasks finished')
                    break

                if i < retries_failed - 1:
                    msg = 'Waiting for retrying failed tasks...'
                    print(msg)
                    _logger.info(msg)
                    time.sleep(60 * 2)

        self._logger.info('Migration exits')
        print('\nexit:', json.dumps(state))
        return state

    def _prepare(self):
        if op.exists(self._shutdown_flag):
            os.unlink(self._shutdown_flag)

        def _handle_term_sig(signum, frame):
            msg = "termination signal received, exiting..."
            print(msg)
            _logger.info(msg)
            with open(self._shutdown_flag, 'w') as f:
                f.write('')

        signal.signal(signal.SIGINT, _handle_term_sig)
        signal.signal(signal.SIGTERM, _handle_term_sig)

    def _discover_tasks(self):
        indexes = self._config.get('indexes')
        data = self._es.search_shards(indexes)
        tasks = []
        for shard in data['shards']:
            for item in shard:
                # Ignore internal index
                if not indexes and item['index'].startswith('.'):
                    continue
                if item['state'] == 'STARTED' and item['primary']:
                    tasks.append(
                        {'es_index': item['index'], 'es_shard': item['shard']},
                    )
        return self._handle_cache(tasks)

    def _handle_cache(self, tasks):
        file_tasks = op.join(self._config.cache_path, 'tasks.json')
        if op.exists(file_tasks):
            with open(file_tasks) as f:
                cont = f.read()
        else:
            cont = '[]'

        try:
            old_tasks = json.loads(cont)
        except json.JSONDecodeError:
            self._logger.error('Invalid task cache', extra={'cache': cont})
            old_tasks = []

        task_map = {
            (task['es_index'], task['es_shard']): task['id']
            for task in old_tasks
        }
        _mappings = IndexLogstoreMappings(
            list([task['es_index'] for task in tasks]),
            self._config.get('logstore_index_mappings'),
        )
        cnt, new_tasks = len(old_tasks), []
        for task in tasks:
            _task = (task['es_index'], task['es_shard'])
            if _task not in task_map:
                task['id'] = cnt
                task['logstore'] = _mappings.get_logstore(task['es_index'])
                new_tasks.append(task)
                cnt += 1
        tasks = old_tasks + new_tasks

        with open(file_tasks, 'w') as f:
            f.write(json.dumps(tasks, indent=2))

        if self._config.get('auto_creation'):
            self._setup_aliyun_log(_mappings)
        return tasks

    def _setup_aliyun_log(self, index_logstore_mappings):
        print('setup aliyun log service...')
        self._logger.info('Setup AliyunLog start')
        logstores = index_logstore_mappings.get_all_logstores()
        for logstore in logstores:
            self._logger.info('Setup AliyunLog', extra={'logstore': logstore})
            self._setup_logstore(index_logstore_mappings, logstore)
        self._logger.info('Init AliyunLog wait')
        time.sleep(self._config.get('wait_time_in_secs'))
        self._logger.info('Init AliyunLog finish')

    def _setup_logstore(self, index_logstore_mappings, logstore):
        try:
            self._log_client.create_logstore(
                project_name=self._config.get('project_name'),
                logstore_name=logstore,
                shard_count=8,
                ttl=3650,
            )
        except LogException as exc:
            if exc.get_error_code() == 'LogStoreAlreadyExist':
                self._logger.info(
                    'Logstore already exist, skip creation.',
                    extra={'logstore': logstore},
                )
            else:
                raise
        self._setup_index(index_logstore_mappings, logstore)

    def _setup_index(self, index_logstore_mappings, logstore):
        indexes = index_logstore_mappings.get_indexes(logstore)
        for index in indexes:
            self._logger.info(
                'Setup AliyunLog Logstore',
                extra={'logstore': logstore, 'es_index': index},
            )
            try:
                resp = self._es.indices.get(index=index)
            except FileNotFoundError:
                self._logger.error('Index not found', extra={'es_index': index})
                continue
            _mappings = resp[index]['mappings']
            if self._es_version >= 7:
                _mappings = {'': _mappings}
            for mappings in _mappings.values():
                index_config = MappingIndexConverter.to_index_config(mappings)
                try:
                    self._log_client.create_index(
                        self._config.get('project_name'),
                        logstore,
                        index_config,
                    )
                except LogException as exc:
                    if exc.get_error_code() == 'IndexAlreadyExist':
                        self._log_client.update_index(
                            self._config.get('project_name'),
                            logstore,
                            index_config,
                        )
                        continue
                    raise


def _migration_worker(config, task, shutdown_flag, es_version):
    if op.exists(shutdown_flag):
        # Already interrupted
        return Checkpoint.interrupted

    extra = {
        'task_id': task['id'],
        'es_index': task['es_index'],
        'es_shard': task['es_shard'],
        'logstore': task['logstore'],
    }
    print('migrate:' + json.dumps(extra))
    logger = PrefixLoggerAdapter('', extra, _logger, {})
    logger.info('Migration worker starts')
    try:
        logstore = MigrationLogstore(
            endpoint=config.get('endpoint'),
            access_id=config.access_key_id,
            access_key=config.access_key,
            project_name=config.get('project_name'),
            logstore_name=task['logstore'],
            topic=config.get('topic'),
            source=config.get('source'),
        )
        task = MigrationTask(
            _id=task['id'],
            es_client=Elasticsearch(
                hosts=config.hosts,
                timeout=60,
                max_retries=30,
                retry_on_timeout=True,
                verify_certs=False,
            ),
            es_index=task['es_index'],
            es_shard=task['es_shard'],
            logstore=logstore,
            ckpt_path=config.ckpt_path,
            time_reference=config.get('time_reference'),
            batch_size=config.get('batch_size'),
            logger=logger,
            es_version=es_version,
            es_query=config.get('query'),
            es_scroll=config.get('scroll'),
        )
        return task.run(shutdown_flag)
    except BaseException:
        logger.error(
            'Exception in migration worker',
            extra=traceback.format_exc(),
        )
        raise
    finally:
        logger.info('Migration worker exits')



# -*- coding: utf-8 -*-

import abc
import logging

from .config import CursorPosition
from ..logexception import LogException
import time
import six
import sys

logger = logging.getLogger(__name__)


class ConsumerProcessorBase(object):
    def __init__(self):
        self.shard_id = -1
        self.last_check_time = 0
        self.checkpoint_timeout = 3

    def save_checkpoint(self, tracker, force=False):
        current_time = time.time()
        if force or current_time - self.last_check_time > self.checkpoint_timeout:
            try:
                self.last_check_time = current_time
                tracker.save_check_point(True)
            except Exception as ex:
                logger.warning(
                    "Fail to store checkpoint for shard {0}, error: {1}".format(self.shard_id, ex))
        else:
            try:
                tracker.save_check_point(False)
            except Exception as ex:
                logger.warning(
                    "Fail to store checkpoint for shard {0}, error: {1}".format(self.shard_id, ex))

    def initialize(self, shard):
        self.shard_id = shard

    @abc.abstractmethod
    def process(self, log_groups, check_point_tracker):
        raise NotImplementedError('not create method process')

    def shutdown(self, check_point_tracker):
        consumer_client = check_point_tracker.consumer_group_client
        _id = '/'.join([
            consumer_client.mproject, consumer_client.mlogstore,
            consumer_client.mconsumer_group, consumer_client.mconsumer,
            str(self.shard_id)
        ])
        logger.info("[%s]ConsumerProcesser is shutdown, shard id: %s", _id,
                    self.shard_id)
        self.save_checkpoint(check_point_tracker, force=True)


class ConsumerProcessorAdaptor(ConsumerProcessorBase):
    def __init__(self, func):
        super(ConsumerProcessorAdaptor, self).__init__()
        self.func = func

    def process(self, log_groups, check_point_tracker):
        ret = self.func(self.shard_id, log_groups)
        if isinstance(ret, bool) and not ret:
            return  # do not save checkpoint when getting False

        self.save_checkpoint(check_point_tracker)


class TaskResult(object):
    def __init__(self, task_exception):
        self.task_exception = task_exception
        self._exc_info = None
        if six.PY2 and task_exception is not None:
            self._exc_info = sys.exc_info()

    def get_exception(self):
        return self.task_exception

    @property
    def exc_info(self):
        if six.PY3:
            return self.task_exception
        else:
            return self._exc_info


class ProcessTaskResult(TaskResult):
    def __init__(self, rollback_check_point):
        super(ProcessTaskResult, self).__init__(None)
        self.rollback_check_point = rollback_check_point

    def get_rollback_check_point(self):
        return self.rollback_check_point


class InitTaskResult(TaskResult):
    def __init__(self, cursor, cursor_persistent, end_cursor):
        super(InitTaskResult, self).__init__(None)
        self.cursor = cursor
        self.cursor_persistent = cursor_persistent
        self._end_cursor = end_cursor

    def get_cursor(self):
        return self.cursor

    @property
    def end_cursor(self):
        return self._end_cursor

    def is_cursor_persistent(self):
        return self.cursor_persistent


class FetchTaskResult(TaskResult):
    def __init__(self, fetched_log_group_list, cursor):
        super(FetchTaskResult, self).__init__(None)
        self.fetched_log_group_list = fetched_log_group_list
        self.cursor = cursor

    def get_fetched_log_group_list(self):
        return self.fetched_log_group_list

    def get_cursor(self):
        return self.cursor


def consumer_process_task(processor, log_groups, check_point_tracker):
    """
    return TaskResult if failed,
    return ProcessTaskResult if succeed
    :param processor:
    :param log_groups:
    :param check_point_tracker:
    :return:
    """
    try:
        check_point = processor.process(log_groups, check_point_tracker)
        check_point_tracker.flush_check()
    except Exception as e:
        return TaskResult(e)
    return ProcessTaskResult(check_point)


def consumer_initialize_task(processor, consumer_client, shard_id, cursor_position, cursor_start_time, cursor_end_time=None):
    """
    return TaskResult if failed, or else, return InitTaskResult
    :param processor:
    :param consumer_client:
    :param shard_id:
    :param cursor_position:
    :param cursor_start_time:
    :return:
    """
    try:
        processor.initialize(shard_id)
        is_cursor_persistent = False
        check_point = consumer_client.get_check_point(shard_id)
        if check_point['checkpoint'] and len(check_point['checkpoint']) > 0:
            is_cursor_persistent = True
            cursor = check_point['checkpoint']
        else:
            if cursor_position == CursorPosition.BEGIN_CURSOR:
                cursor = consumer_client.get_begin_cursor(shard_id)
            elif cursor_position == CursorPosition.END_CURSOR:
                cursor = consumer_client.get_end_cursor(shard_id)
            else:
                cursor = consumer_client.get_cursor(shard_id, cursor_start_time)

        end_cursor = None
        if cursor_end_time is not None:
            end_cursor = consumer_client.get_cursor(shard_id, cursor_end_time)

        return InitTaskResult(cursor, is_cursor_persistent, end_cursor)
    except Exception as e:
        return TaskResult(e)


def consumer_fetch_task(loghub_client_adapter, shard_id, cursor, max_fetch_log_group_size=1000, end_cursor=None):
    exception = None

    for retry_times in range(3):
        try:
            response = loghub_client_adapter.pull_logs(shard_id, cursor, count=max_fetch_log_group_size, end_cursor=end_cursor)
            fetch_log_group_list = response.get_loggroup_list()
            next_cursor = response.get_next_cursor()
            logger.debug("shard id = %s cursor = %s next cursor = %s size: %s",
                         shard_id, cursor, next_cursor,
                         response.get_log_count())
            if not next_cursor:
                return FetchTaskResult(fetch_log_group_list, cursor)
            else:
                return FetchTaskResult(fetch_log_group_list, next_cursor)
        except LogException as e:
            exception = e
        except Exception as e1:
            logger.error(e1, exc_info=True)
            raise Exception(e1)

        # only retry if the first request get "SLSInvalidCursor" exception
        if retry_times == 0 and isinstance(exception, LogException) \
                and 'invalidcursor' in exception.get_error_code().lower():
            try:
                cursor = loghub_client_adapter.get_end_cursor(shard_id)
            except Exception:
                return TaskResult(exception)
        else:
            break

    return TaskResult(exception)


def consumer_shutdown_task(processor, check_point_tracker):
    """
    :param processor:
    :param check_point_tracker:
    :return:
    """
    exception = None
    try:
        processor.shutdown(check_point_tracker)
    except Exception as e:
        print(e)
        exception = None

    try:
        check_point_tracker.get_check_point()
    except Exception:
        logger.error('Failed to flush check point', exc_info=True)

    return TaskResult(exception)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

import json


class LogException(Exception):
    """The Exception of the log request & response.
    
    :type errorCode: string
    :param errorCode: log service error code 
    
    :type errorMessage: string
    :param errorMessage: detailed information for the exception
    
    :type requestId: string
    :param requestId: the request id of the response, '' is set if client error
    """

    def __init__(self, errorCode, errorMessage, requestId='', resp_status=200, resp_header='', resp_body=''):
        self._errorCode = errorCode
        self._errorMessage = errorMessage
        self._requestId = requestId
        self.resp_status = resp_status
        self.resp_header = resp_header
        self.resp_body = resp_body

        if not self.resp_body:
            self.resp_body = json.dumps({
                "errorCode": self._errorCode,
                "errorMessage": self._errorMessage,
                "requestId": self._requestId
            })

    def __str__(self):
        return json.dumps({
            "errorCode": self._errorCode,
            "errorMessage": self._errorMessage,
            "requestId": self._requestId
        }, sort_keys=True)

    def get_error_code(self):
        """ return error code of exception
        
        :return: string, error code of exception.
        """
        return self._errorCode

    def get_error_message(self):
        """ return error message of exception
        
        :return: string, error message of exception.
        """
        return self._errorMessage

    def get_request_id(self):
        """ return request id of exception. if client exception, request id is empty string
        
        :return: string, request id of exception.
        """
        return self._requestId

    def get_resp_body(self):
        return self.resp_body



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

__all__ = ['CreateLogStoreResponse', 'DeleteLogStoreResponse', 'GetLogStoreResponse',
           'UpdateLogStoreResponse', 'ListLogStoreResponse']

from .util import Util
from .logresponse import LogResponse


class CreateLogStoreResponse(LogResponse):
    """ The response of the create_logstore API from log.
    
    :type header: dict
    :param header: CreateLogStoreResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('CreateLogStoreResponse:')
        print('headers:', self.get_all_headers())


class DeleteLogStoreResponse(LogResponse):
    """ The response of the delete_logstore API from log.
    
    :type header: dict
    :param header: DeleteLogStoreResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('DeleteLogStoreResponse:')
        print('headers:', self.get_all_headers())


class GetLogStoreResponse(LogResponse):
    """ The response of the get_logstore API from log.
    
    :type header: dict
    :param header: GetLogStoreResponse HTTP response header

    :type resp: dict
    :param resp: the HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.logstore_name = Util.convert_unicode_to_str(resp["logstoreName"])
        self.ttl = int(resp["ttl"])
        self.shard_count = int(resp["shardCount"])
        self.enable_tracking = bool(resp["enable_tracking"])
        self.append_meta = bool(resp["appendMeta"])
        self.auto_split = bool(resp["autoSplit"])
        self.max_split_shard = int(resp["maxSplitShard"])
        self.preserve_storage = self.ttl >= 3650
        self.encrypt_conf = None
        if 'encrypt_conf' in resp:
            self.encrypt_conf = resp["encrypt_conf"]

    def get_shard_count(self):
        """

        :return:
        """
        return self.shard_count

    def get_ttl(self):
        """

        :return:
        """
        return self.ttl

    def get_enable_tracking(self):
        """

        :return:
        """
        return self.enable_tracking

    def get_encrypt_conf(self):
        """

        :return:
        """
        return self.encrypt_conf

    def log_print(self):
        """

        :return:
        """
        print('GetLogStoreResponse:')
        print('headers:', self.get_all_headers())
        print('logstore_name:', self.logstore_name)
        print('shard_count:', str(self.shard_count))
        print('ttl:', str(self.ttl))
        if self.encrypt_conf != None:
            print('encrypt_conf:', str(self.encrypt_conf))


class UpdateLogStoreResponse(LogResponse):
    """ The response of the update_logstore API from log.
    
    :type header: dict
    :param header: UpdateLogStoreResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpdateLogStoreResponse:')
        print('headers:', self.get_all_headers())


class ListLogStoreResponse(LogResponse):
    """ The response of the list_logstore API from log.
    
    :type header: dict
    :param header: ListLogStoreResponse HTTP response header

    :type resp: dict
    :param resp: the HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = int(resp["count"])
        self.total_count = int(resp["total"])
        self._logstores = Util.convert_unicode_to_str(resp.get("logstores", []))

    def get_logstores(self):
        """

        :return:
        """
        return self.logstores

    def get_count(self):
        return self.count

    def get_logstores_count(self):
        """

        :return:
        """
        return self.count

    def get_logstores_total(self):
        """

        :return:
        """
        return self.total_count

    def get_total(self):
        """

        :return:
        """
        return self.total_count

    def log_print(self):
        """

        :return:
        """
        print('ListLogStoreResponse:')
        print('headers:', self.get_all_headers())
        print('logstores_count:', str(self.count))
        print('logstores_total:', str(self.total_count))
        print('logstores:', str(self._logstores))

    def merge(self, response):
        if not isinstance(response, ListLogStoreResponse):
            raise ValueError("passed response is not a ListLogstoresResponse: " + str(type(response)))

        self.count += response.get_count()
        self.total_count = response.get_total() # use the latest total count
        self.logstores.extend(response.get_logstores())

        # update body
        self.body = {
            'count': self.count,
            'total': self.total,
            'logstores': self.logstores
        }

        return self

    @property
    def total(self,):
        return self.total_count

    @property
    def logstores(self):
        return self._logstores



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .util import Util


class LogResponse(object):
    """ The base response class of all log response. 
    
    :type headers: dict
    :param headers: HTTP response header
    """

    def __init__(self, headers, body=''):
        self.headers = headers
        if body is None:
            body = ''
        self.body = body

    def get_request_id(self):
        """ Get the request id of the response.  '' will be return if not set.
        
        :return: string, request id
        """
        return Util.h_v_td(self.headers, 'x-log-requestid', '')

    def get_body(self):
        """ Get body

        :return: string
        """
        return self.body

    def get_all_headers(self):
        """ Get all http header of the response
        
        :return: dict, response header
        """
        return self.headers

    def get_header(self, key):
        """ Get specified http header of the response, '' will be return if not set.

        :type key: string
        :param key: the key to get header
        
        :return: string, response header
        """
        return self.headers[key] if key in self.headers else ''

    def log_print(self):
        print('header: ', self.headers)



"""
syslog RFC
https://datatracker.ietf.org/doc/rfc5424/
https://tools.ietf.org/html/rfc3164
"""
import socket
from datetime import datetime
import ssl
import six


def datetime2rfc3339(dt, is_utc=False):
    if not is_utc:
        # calculating timezone
        d1 = datetime.now()
        d2 = datetime.utcnow()
        diff_hr = (d1 - d2).seconds / 60 / 60
        tz = ""

        if diff_hr == 0:
            tz = "Z"
        else:
            if diff_hr > 0:
                tz = "+%s" % (tz)

            tz = "%s%.2d%.2d" % (tz, diff_hr, 0)

        return "%s%s" % (dt.strftime("%Y-%m-%dT%H:%M:%S.%f"), tz)

    else:
        return dt.isoformat() + 'Z'


FAC_KERNEL = 0
FAC_USER = 1
FAC_MAIL = 2
FAC_SYSTEM = 3
FAC_SECURITY = 4
FAC_SYSLOG = 5
FAC_PRINTER = 6
FAC_NETWORK = 7
FAC_UUCP = 8
FAC_CLOCK = 9
FAC_AUTH = 10
FAC_FTP = 11
FAC_NTP = 12
FAC_LOG_AUDIT = 13
FAC_LOG_ALERT = 14
FAC_CLOCK2 = 15
FAC_LOCAL0 = 16
FAC_LOCAL1 = 17
FAC_LOCAL2 = 18
FAC_LOCAL3 = 19
FAC_LOCAL4 = 20
FAC_LOCAL5 = 21
FAC_LOCAL6 = 22
FAC_LOCAL7 = 23

SEV_EMERGENCY = 0
SEV_ALERT = 1
SEV_CRITICAL = 2
SEV_ERROR = 3
SEV_WARNING = 4
SEV_NOTICE = 5
SEV_INFO = 6
SEV_DEBUG = 7


class SyslogClientBase(object):
    def __init__(self, server, port, proto='udp', clientname=None,
                 maxMessageLength=1024, timeout=120, cert_path=None):
        self.socket = None
        self.server = server
        self.port = port
        self.proto = socket.SOCK_DGRAM
        self.ssl_kwargs = None

        self.maxMessageLength = maxMessageLength
        self.timeout = timeout

        if proto is not None:
            if proto.upper() == 'UDP':
                self.proto = socket.SOCK_DGRAM
            elif proto.upper() == 'TCP':
                self.proto = socket.SOCK_STREAM
            elif proto.upper() == 'TLS':
                self.proto = socket.SOCK_STREAM
                self.ssl_kwargs ={
                                    'cert_reqs': ssl.CERT_REQUIRED,
                                    'ssl_version': ssl.PROTOCOL_TLS,
                                    'ca_certs': cert_path,
                                  }

        self.clientname = clientname or socket.getfqdn() or socket.gethostname()
        self.cert_path = cert_path

    def connect(self):
        if self.socket is None:
            r = socket.getaddrinfo(self.server, self.port, socket.AF_UNSPEC, self.proto)
            if r is None:
                return False
            for (addr_fam, sock_kind, proto, ca_name, sock_addr) in r:

                sock = socket.socket(addr_fam, self.proto)
                if six.PY3 and self.ssl_kwargs:
                    self.socket = ssl.wrap_socket(sock, **self.ssl_kwargs)
                else:
                    self.socket = sock

                if self.socket is None:
                    return False

                self.socket.settimeout(self.timeout)

                try:
                    self.socket.connect(sock_addr)
                    return True

                except socket.timeout as e:
                    if self.socket is not None:
                        self.socket.close()
                        self.socket = None
                    continue

                except socket.error as e:
                    if self.socket is not None:
                        self.socket.close()
                        self.socket = None
                    continue

            return False

        else:
            return True

    def __enter__(self):
        if not self.connect():
            raise ValueError("cannot connect to remote syslog server")

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self.socket is not None:
            self.socket.close()
            self.socket = None

    def log(self, message, timestamp=None, hostname=None, facility=None,
            severity=None):
        pass

    def send(self, messagedata):
        if self.socket is not None or self.connect():
            try:
                if self.maxMessageLength is not None:
                    self.socket.sendall(messagedata[:self.maxMessageLength])
                else:
                    self.socket.sendall(messagedata)
            except IOError as e:
                self.close()


class SyslogClientRFC5424(SyslogClientBase):
    def __init__(self, server, port, proto='udp',
                 clientname=None, timeout=120, cert_path=None):
        SyslogClientBase.__init__(self,
                                  server=server,
                                  port=port,
                                  proto=proto,
                                  clientname=clientname,
                                  maxMessageLength=None,
                                  timeout=timeout,
                                  cert_path=cert_path
                                  )

    def log(self, message, facility=None, severity=None, timestamp=None,
            hostname=None, version=1, program=None, pid=None, msgid=None):

        facility = FAC_USER if facility is None else facility
        severity = SEV_INFO if severity is None else severity
        pri = facility * 8 + severity
        hostname_s = hostname or self.clientname
        appname_s = program or "-"
        procid_s = pid or "-"
        msgid_s = msgid or "-"

        if timestamp is None:
            timestamp_s = datetime2rfc3339(datetime.utcnow(), is_utc=True)
        else:
            timestamp_s = datetime2rfc3339(timestamp, is_utc=False)

        d = "<%i>%i %s %s %s %s %s %s\n" % (
            pri,
            version,
            timestamp_s,
            hostname_s,
            appname_s,
            procid_s,
            msgid_s,
            message
        )

        self.send(d.encode('utf-8', errors='ignore'))


class SyslogClientRFC3164(SyslogClientBase):
    def __init__(self, server, port, proto='udp', clientname=None, timeout=120, cert_path=None):
        SyslogClientBase.__init__(self,
                                  server=server,
                                  port=port,
                                  proto=proto,
                                  clientname=clientname,
                                  maxMessageLength=1024,
                                  timeout=timeout,
                                  cert_path=cert_path
                                  )

    def log(self, message, facility=None, severity=None, timestamp=None,
            hostname=None, program="SyslogClient", pid=None):
        facility = FAC_USER if facility is None else facility
        severity = SEV_INFO if severity is None else severity

        pri = facility * 8 + severity

        t = timestamp or datetime.now()

        timestamp_s = t.strftime("%b %d %H:%M:%S")

        hostname_s = hostname or self.clientname

        tag_s = ""
        if program is None:
            tag_s += "SyslogClient"
        else:
            tag_s += program

        if pid is not None:
            tag_s += "[%i]" % (pid,)

        d = "<%i>%s %s %s: %s\n" % (
            pri,
            timestamp_s,
            hostname_s,
            tag_s,
            message
        )

        self.send(d.encode('ASCII', errors='ignore'))



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logexception import LogException
from .logresponse import LogResponse
from .util import Util
from .util import base64_encodestring as b64e

from .log_logs_pb2 import LogGroupList
from .log_logs_raw_pb2 import LogGroupListRaw
import six

DEFAULT_DECODE_LIST = ('utf8',)


class PullLogResponse(LogResponse):
    """ The response of the pull_logs API from log.
    
    :type header: dict
    :param header: PullLogResponse HTTP response header

    :type resp: string
    :param resp: the HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self._is_bytes_type = None
        self.next_cursor = Util.convert_unicode_to_str(Util.h_v_t(header, "x-log-cursor"))
        self.log_count = int(Util.h_v_t(header, "x-log-count"))
        self.loggroup_list = LogGroupList()

        self._parse_loggroup_list(resp)
        self.loggroup_list_json = None
        self.flatten_logs_json = None
        self._body = None

    def get_body(self):
        if self._body is None:
            self._body = {"next_cursor": self.next_cursor,
                          "count": len(self.get_flatten_logs_json()),
                          "logs": self.get_flatten_logs_json()}
        return self._body

    @property
    def body(self):
        return self.get_body()

    @body.setter
    def body(self, value):
        self._body = value

    def get_next_cursor(self):
        return self.next_cursor

    def get_log_count(self):
        return len(self.get_flatten_logs_json())

    def get_loggroup_count(self):
        return len(self.loggroup_list.LogGroups)

    def get_loggroup_json_list(self):
        if self.loggroup_list_json is None:
            self._transfer_to_json()
        return self.loggroup_list_json

    def get_loggroup_list(self):
        return self.loggroup_list

    def get_loggroup(self, index):
        if index < 0 or index >= len(self.loggroup_list.LogGroups):
            return None
        return self.loggroup_list.LogGroups[index]

    def log_print(self):
        print('PullLogResponse')
        print('next_cursor', self.next_cursor)
        print('log_count', self.log_count)
        print('headers:', self.get_all_headers())
        print('detail:', self.get_loggroup_json_list())

    def _parse_loggroup_list(self, data):
        try:
            self.loggroup_list.ParseFromString(data)
        except Exception as ex:
            ex_second = None
            try:
                p = LogGroupListRaw()
                p.ParseFromString(data)
                self.loggroup_list = p
                self._is_bytes_type = True
                return
            except Exception as ex2:
                ex_second = ex2

            err = 'failed to parse data to LogGroupList: \n' \
                  + str(ex) + '\n' + str(ex_second) + '\nb64 raw data:\n' + b64e(data) \
                  + '\nheader:' + str(self.headers)
            raise LogException('BadResponse', err)

    def _transfer_to_json(self):
        self.loggroup_list_json = []
        for logGroup in self.loggroup_list.LogGroups:
            items = []
            tags = {}
            for tag in logGroup.LogTags:
                tags[tag.Key] = tag.Value

            for log in logGroup.Logs:
                item = {'@lh_time': log.Time}
                for content in log.Contents:
                    item[content.Key] = content.Value
                items.append(item)
            log_items = {'topic': logGroup.Topic, 'source': logGroup.Source,
                         'logs': items,
                         'tags': tags}
            self.loggroup_list_json.append(log_items)


    @staticmethod
    def _b2u(content):
        if six.PY3 and isinstance(content, six.binary_type):
            if len(DEFAULT_DECODE_LIST) == 1:
                return content.decode(DEFAULT_DECODE_LIST[0], 'ignore')

            for d in DEFAULT_DECODE_LIST:
                try:
                    return content.decode(d)
                except Exception as ex:
                    continue

            # force to use utf8
            return content.decode('utf8', 'ignore')
        return content

    @staticmethod
    def get_log_count_from_group(loggroup_list):
        count = 0
        for logGroup in loggroup_list.LogGroups:
            for log in logGroup.Logs:
                count += 1
        return count

    @staticmethod
    def loggroups_to_flattern_list(loggroup_list, time_as_str=None, decode_bytes=None):
        flatten_logs_json = []
        for logGroup in loggroup_list.LogGroups:
            tags = {}
            for tag in logGroup.LogTags:
                tags[u"__tag__:{0}".format(tag.Key)] = tag.Value

            for log in logGroup.Logs:
                item = {u'__time__': six.text_type(log.Time) if time_as_str else log.Time,
                        u'__topic__': logGroup.Topic,
                        u'__source__': logGroup.Source}
                item.update(tags)
                for content in log.Contents:
                    item[PullLogResponse._b2u(content.Key) if decode_bytes else content.Key] = PullLogResponse._b2u(content.Value) if decode_bytes else content.Value
                flatten_logs_json.append(item)
        return flatten_logs_json

    def get_flatten_logs_json(self, time_as_str=None, decode_bytes=None):
        decode_bytes = decode_bytes or self._is_bytes_type

        if self.flatten_logs_json is None:
            self.flatten_logs_json = self.loggroups_to_flattern_list(self.loggroup_list, time_as_str=time_as_str,
                                                                     decode_bytes=decode_bytes)

        return self.flatten_logs_json

    def get_flatten_logs_json_auto(self):
        if self.flatten_logs_json is None:
            self.flatten_logs_json = self.loggroups_to_flattern_list(self.loggroup_list, time_as_str=True,
                                                                     decode_bytes=self._is_bytes_type)

        return self.flatten_logs_json



import ast
import inspect
import logging
try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable

import six
from functools import wraps

from .etl_util import re_full_match
from .settings import TransFnType, check_fn_type_by_name, builtin_macros

logger = logging.getLogger(__name__)


class ConfigParser(ast.NodeVisitor):
    @staticmethod
    def get_fn(fn, name):
        fn_type = check_fn_type_by_name(name)
        if fn_type == TransFnType.EVENT_NEW:
            return fn
        if fn_type == TransFnType.EVENT_UPDATE:
            @wraps(fn)
            def new_event_fn(event):
                r = fn(event)
                event.update(r)
                return event

            return new_event_fn

        return None

    @staticmethod
    def macro_check(name):
        for k, v in six.iteritems(builtin_macros):
            if re_full_match(k, name):
                return v

    def __init__(self, md):
        self.name_list = []
        self.fn_list = []
        self.md = md

    def visit_Name(self, node):
        if hasattr(self.md, node.id):
            obj = getattr(self.md, node.id)
            if isinstance(obj, Callable):
                fn = self.get_fn(obj, node.id)
                if fn:
                    logging.info("get name {0} which is a fn actually, add it".format(node.id))
                    self.name_list.append([node.lineno, fn])
            else:
                fn_map = self.macro_check(node.id)
                if fn_map is not None:
                    v = getattr(self.md, node.id)
                    self.name_list.append([node.lineno, fn_map(v)])
                else:
                    logging.info("get name {0} not in macro list, skip it".format(node.id))

    def generic_visit(self, node):
        if isinstance(node, ast.FunctionDef):
            obj = getattr(self.md, node.name, None)
            if inspect.isfunction(obj):
                fn = self.get_fn(obj, node.name)
                if fn:
                    self.fn_list.append([node.lineno, fn])

        ast.NodeVisitor.generic_visit(self, node)

    def parse(self):
        code = inspect.getsource(self.md)

        self.visit(ast.parse(code))
        return sorted(self.name_list + self.fn_list, key=lambda x: x[0])




import logging
import re
import six

from .trans_base import trans_comp_check_mdoe_base
from ..exceptions import SettingError

logger = logging.getLogger(__name__)

__all__ = ['trans_comp_regex']


class trans_comp_regex(trans_comp_check_mdoe_base):
    """
    # full match: ^$
    # ignore group: (?:...)
    # case insensitive: (?i)
    """

    def __init__(self, pattern, fields_info=None, mode=None):
        super(trans_comp_regex, self).__init__(mode=mode)
        pattern = self._u(pattern)
        fields_info = self._u(fields_info)

        self.config = pattern

        try:
            self.ptn = re.compile(pattern)
        except re.error as ex:
            logger.error(u'transform_regex: cannot compile pattern: "{0}"'.format(ex))
            raise SettingError(ex, pattern)

        self.fields_info = None
        if isinstance(fields_info, (six.text_type, six.binary_type, list, dict)):
            self.fields_info = fields_info
        elif fields_info is not None:
            logger.warning(u'transform_regex: unknown fields info type: "{0}"'.format(fields_info))

    def __call__(self, event, inpt):
        inpt = self._u(inpt)

        if not isinstance(inpt, list):
            inputs = [inpt]
        else:
            inputs = inpt

        for data in inputs:
            if data in event:
                m = None
                find_iter = iter(self.ptn.finditer(event[data]))
                for x in find_iter:
                    m = x  # only find the first one cause there's no field info, no meaning to find others.
                    break

                if m:
                    self.sets(event, m.groupdict())
                else:
                    logger.info(u'transform_regex: field value "{0}" cannot extract value with config "{1}"'
                                .format(event[data], self.config))

                if self.fields_info:
                    # use m first
                    if not m:
                        logger.warning(
                            u'transform_regex: field value "{0}" cannot extract value '
                            u'with config "{1}" with field info {2}'.format(event[data], self.config, self.fields_info))
                        break

                    if isinstance(self.fields_info, (six.binary_type, six.text_type)):
                        # only find first one
                        k, v = self.fields_info, m.group()
                        self.set(event, k, v)
                    elif isinstance(self.fields_info, (dict, )):
                        ms = [m] + list(find_iter)

                        for i, m in enumerate(ms):
                            for k, v in six.iteritems(self.fields_info):
                                try:
                                    kk, vv = m.expand(k), m.expand(v)
                                    self.set(event, kk, vv, check_kw_name=True)
                                except re.error as ex:
                                    logger.info(
                                        u'transform_regex: cannot expand matched group "{0}" for fields info "{1}:{2}", '
                                        u'detail: "{3}"'.format(
                                            m.group(), k, v, ex
                                        ))
                    elif isinstance(self.fields_info, list):
                        if m.groups():  # there's group, use group mode
                            for i, g in enumerate(m.groups()):
                                if i >= len(self.fields_info):
                                    logger.warning(
                                        u'transform_regex: field value "{0}" captured group count not equal to '
                                        u'config "{1}" with field info {2}'.format(event[data],
                                                                                  self.config, self.fields_info))
                                    break

                                k, v = self.fields_info[i], g
                                self.set(event, k, v, check_kw_name=True)
                        else:
                            ms = [m] + list(find_iter)
                            if len(ms) != len(self.fields_info):
                                logger.warning(
                                    u'transform_regex: field value "{0}" match count not equal to '
                                    u'config "{1}" with field info {2}'.format(event[data], self.config,
                                                                              self.fields_info))
                            for i, m in enumerate(ms):
                                if i >= len(self.fields_info):
                                    break

                                k, v = self.fields_info[i], m.group()
                                self.set(event, k, v, check_kw_name=True)
                    else:
                        logger.warning(u'transform_regex: unknown fields info type: "{0}"'.format(self.fields_info))
            else:
                logger.info(u'transform_regex: event "{0}" doesn not contain field "{1}"'.format(event, data))

        return event




from .logclient import LogClient
from .logexception import LogException
from .gethistogramsrequest import GetHistogramsRequest
from .getlogsrequest import GetLogsRequest, GetProjectLogsRequest
from .index_config import IndexConfig, IndexKeyConfig, IndexLineConfig
from .listtopicsrequest import ListTopicsRequest
from .listlogstoresrequest import ListLogstoresRequest
from .logtail_config_detail import *
from .machine_group_detail import MachineGroupDetail
from .putlogsrequest import PutLogsRequest
from .shipper_config import ShipperTask, OssShipperConfig, OdpsShipperConfig
from .version import __version__
from .logitem import LogItem
from .consumer_group_request import *
from .external_store_config import * 

# response class
from .consumer_group_response import *
from .cursor_response import GetCursorResponse
from .cursor_time_response import GetCursorTimeResponse
from .gethistogramsresponse import GetHistogramsResponse, Histogram
from .getlogsresponse import GetLogsResponse, QueriedLog
from .index_config_response import *
from .listlogstoresresponse import ListLogstoresResponse
from .listtopicsresponse import ListTopicsResponse
from .logresponse import LogResponse
from .logtail_config_response import *
from .machinegroup_response import *
from .project_response import *
from .pulllog_response import PullLogResponse
from .shard_response import *
from .shipper_response import *
from .common_response import *
from .external_store_config_response import * 
from .log_logs_raw_pb2 import LogGroupRaw as LogGroup

# logging handler
from .logger_hanlder import SimpleLogHandler, QueuedLogHandler, LogFields, UwsgiQueuedLogHandler



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: log_logs_raw.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='log_logs_raw.proto',
  package='',
  syntax='proto2',
  serialized_pb=_b('\n\x12log_logs_raw.proto\"`\n\x06LogRaw\x12\x0c\n\x04Time\x18\x01 \x02(\r\x12!\n\x08\x43ontents\x18\x02 \x03(\x0b\x32\x0f.LogRaw.Content\x1a%\n\x07\x43ontent\x12\x0b\n\x03Key\x18\x01 \x02(\t\x12\r\n\x05Value\x18\x02 \x02(\x0c\"\'\n\tLogTagRaw\x12\x0b\n\x03Key\x18\x01 \x02(\t\x12\r\n\x05Value\x18\x02 \x02(\t\"\x87\x01\n\x0bLogGroupRaw\x12\x15\n\x04Logs\x18\x01 \x03(\x0b\x32\x07.LogRaw\x12\x10\n\x08Reserved\x18\x02 \x01(\t\x12\r\n\x05Topic\x18\x03 \x01(\t\x12\x0e\n\x06Source\x18\x04 \x01(\t\x12\x13\n\x0bMachineUUID\x18\x05 \x01(\t\x12\x1b\n\x07LogTags\x18\x06 \x03(\x0b\x32\n.LogTagRaw\"2\n\x0fLogGroupListRaw\x12\x1f\n\tLogGroups\x18\x01 \x03(\x0b\x32\x0c.LogGroupRaw')
)




_LOGRAW_CONTENT = _descriptor.Descriptor(
  name='Content',
  full_name='LogRaw.Content',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Key', full_name='LogRaw.Content.Key', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='Value', full_name='LogRaw.Content.Value', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=81,
  serialized_end=118,
)

_LOGRAW = _descriptor.Descriptor(
  name='LogRaw',
  full_name='LogRaw',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Time', full_name='LogRaw.Time', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='Contents', full_name='LogRaw.Contents', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_LOGRAW_CONTENT, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=118,
)


_LOGTAGRAW = _descriptor.Descriptor(
  name='LogTagRaw',
  full_name='LogTagRaw',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Key', full_name='LogTagRaw.Key', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='Value', full_name='LogTagRaw.Value', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=120,
  serialized_end=159,
)


_LOGGROUPRAW = _descriptor.Descriptor(
  name='LogGroupRaw',
  full_name='LogGroupRaw',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Logs', full_name='LogGroupRaw.Logs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='Reserved', full_name='LogGroupRaw.Reserved', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='Topic', full_name='LogGroupRaw.Topic', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='Source', full_name='LogGroupRaw.Source', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='MachineUUID', full_name='LogGroupRaw.MachineUUID', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='LogTags', full_name='LogGroupRaw.LogTags', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=162,
  serialized_end=297,
)


_LOGGROUPLISTRAW = _descriptor.Descriptor(
  name='LogGroupListRaw',
  full_name='LogGroupListRaw',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='LogGroups', full_name='LogGroupListRaw.LogGroups', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=299,
  serialized_end=349,
)

_LOGRAW_CONTENT.containing_type = _LOGRAW
_LOGRAW.fields_by_name['Contents'].message_type = _LOGRAW_CONTENT
_LOGGROUPRAW.fields_by_name['Logs'].message_type = _LOGRAW
_LOGGROUPRAW.fields_by_name['LogTags'].message_type = _LOGTAGRAW
_LOGGROUPLISTRAW.fields_by_name['LogGroups'].message_type = _LOGGROUPRAW
DESCRIPTOR.message_types_by_name['LogRaw'] = _LOGRAW
DESCRIPTOR.message_types_by_name['LogTagRaw'] = _LOGTAGRAW
DESCRIPTOR.message_types_by_name['LogGroupRaw'] = _LOGGROUPRAW
DESCRIPTOR.message_types_by_name['LogGroupListRaw'] = _LOGGROUPLISTRAW
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LogRaw = _reflection.GeneratedProtocolMessageType('LogRaw', (_message.Message,), dict(

  Content = _reflection.GeneratedProtocolMessageType('Content', (_message.Message,), dict(
    DESCRIPTOR = _LOGRAW_CONTENT,
    __module__ = 'log_logs_raw_pb2'
    # @@protoc_insertion_point(class_scope:LogRaw.Content)
    ))
  ,
  DESCRIPTOR = _LOGRAW,
  __module__ = 'log_logs_raw_pb2'
  # @@protoc_insertion_point(class_scope:LogRaw)
  ))
_sym_db.RegisterMessage(LogRaw)
_sym_db.RegisterMessage(LogRaw.Content)

LogTagRaw = _reflection.GeneratedProtocolMessageType('LogTagRaw', (_message.Message,), dict(
  DESCRIPTOR = _LOGTAGRAW,
  __module__ = 'log_logs_raw_pb2'
  # @@protoc_insertion_point(class_scope:LogTagRaw)
  ))
_sym_db.RegisterMessage(LogTagRaw)

LogGroupRaw = _reflection.GeneratedProtocolMessageType('LogGroupRaw', (_message.Message,), dict(
  DESCRIPTOR = _LOGGROUPRAW,
  __module__ = 'log_logs_raw_pb2'
  # @@protoc_insertion_point(class_scope:LogGroupRaw)
  ))
_sym_db.RegisterMessage(LogGroupRaw)

LogGroupListRaw = _reflection.GeneratedProtocolMessageType('LogGroupListRaw', (_message.Message,), dict(
  DESCRIPTOR = _LOGGROUPLISTRAW,
  __module__ = 'log_logs_raw_pb2'
  # @@protoc_insertion_point(class_scope:LogGroupListRaw)
  ))
_sym_db.RegisterMessage(LogGroupListRaw)


# @@protoc_insertion_point(module_scope)



from .transform_list import transform
from .condition_list import condition
from functools import wraps
from .transform_base import transform_base
from ..etl_util import process_event, support_event_list_simple
import logging

__all__ = ['dispatch_event', 'transform_event', 'drop_event', 'keep_event']

logger = logging.getLogger(__name__)


class dispatch_event(transform_base):
    def __init__(self, config):
        if not isinstance(config, list):
            self.trans = [config]
        else:
            self.trans = config

        self.trans = [(condition(c), transform(t)) for c, t in self.trans]

    @support_event_list_simple
    def __call__(self, event):
        for c, t in self.trans:
            if c(event):  # only match once
                return t(event)

        return event


# this function is used to bypass lambda trap
def _get_ct(c, t):
    @wraps(t)
    def _fn(e):
        if c(e):
            return t(e)
        return e
    return _fn


class transform_event(transform_base):
    def __init__(self, config):
        if not isinstance(config, list):
            trans = [config]
        else:
            trans = config

        self.fn_list = [_get_ct(condition(c), transform(t)) for c, t in trans]

    def __call__(self, event):
        return process_event(event, self.fn_list)


class drop_event(transform_base):
    def __init__(self, cond):
        self.c = condition(cond)

    @support_event_list_simple
    def __call__(self, event):
        if self.c(event):
            logger.info(u"drop_event: drop event: {0}".format(event))
            return None
        return event


class keep_event(transform_base):
    def __init__(self, cond):
        self.c = condition(cond)

    @support_event_list_simple
    def __call__(self, event):
        if self.c(event):
            return event
        logger.info(u"keep_event: drop event: {0}".format(event))
        return None





# -*- coding: utf-8 -*-


from .logresponse import LogResponse
import six
import json

__all__ = ['ConsumerGroupEntity',
           'ConsumerGroupCheckPointResponse',
           'ConsumerGroupHeartBeatResponse',
           'ConsumerGroupUpdateCheckPointResponse',
           'CreateConsumerGroupResponse',
           'DeleteConsumerGroupResponse',
           'ListConsumerGroupResponse',
           'UpdateConsumerGroupResponse']


class ConsumerGroupEntity(object):
    def __init__(self, consumer_group_name, timeout, in_order=False):
        self.consumer_group_name = consumer_group_name
        self.timeout = timeout
        self.in_order = in_order

    def get_consumer_group_name(self):
        """

        :return:
        """
        return self.consumer_group_name

    def set_consumer_group_name(self, consumer_group_name):
        """

        :param consumer_group_name:
        :return:
        """
        self.consumer_group_name = consumer_group_name

    def get_timeout(self):
        """

        :return:
        """
        return self.timeout

    def set_timeout(self, timeout):
        """

        :param timeout:
        :return:
        """
        self.timeout = timeout

    def is_in_order(self):
        """

        :return:
        """
        return self.in_order

    def set_in_order(self, in_order):
        """

        :param in_order:
        :return:
        """
        self.in_order = in_order

    def to_request_json(self):
        """

        :return:
        """
        log_store_dict = {
            'consumerGroup': self.get_consumer_group_name(),
            'timeout': self.get_timeout(),
            'order': self.is_in_order()
        }
        return six.b(json.dumps(log_store_dict))

    def to_string(self):
        return "ConsumerGroup [consumerGroupName=" + self.consumer_group_name \
               + ", timeout=" + str(self.timeout) + ", inOrder=" + str(self.in_order) + "]"


class CreateConsumerGroupResponse(LogResponse):
    def __init__(self, headers, resp=''):
        LogResponse.__init__(self, headers, resp)


class ConsumerGroupCheckPointResponse(LogResponse):
    def __init__(self, resp, headers):
        LogResponse.__init__(self, headers, resp)
        self.count = len(resp)
        self.consumer_group_check_poins = resp

    def get_count(self):
        """

        :return:
        """
        return self.count

    def get_consumer_group_check_points(self):
        """

        :return:
        """
        return self.consumer_group_check_poins

    def log_print(self):
        """

        :return:
        """
        print('ListConsumerGroupCheckPoints:')
        print('headers:', self.get_all_headers())
        print('count:', self.count)
        print('consumer_group_check_points:', self.consumer_group_check_poins)

    def check_checkpoint(self, client, project_name, logstore_name):
        """

        :param client:
        :param project_name:
        :param logstore_name:
        :return:
        """
        for checkpoint in self.consumer_group_check_poins:
            cursor = checkpoint["checkpoint"]
            shard_id = checkpoint["shard"]
            if cursor:
                ret = client.get_previous_cursor_time(project_name, logstore_name, shard_id, cursor)
                checkpoint["checkpoint_previous_cursor_time"] = ret.get_cursor_time()


class ConsumerGroupHeartBeatResponse(LogResponse):
    def __init__(self, resp, headers):
        LogResponse.__init__(self, headers, resp)
        self.shards = resp

    def get_shards(self):
        """

        :return:
        """
        return self.shards

    def set_shards(self, shards):
        """

        :param shards:
        :return:
        """
        self.shards = shards

    def log_print(self):
        """

        :return:
        """
        print('ListHeartBeat:')
        print('headers:', self.get_all_headers())
        print('shards:', self.shards)


class ConsumerGroupUpdateCheckPointResponse(LogResponse):
    def __init__(self, headers, resp=''):
        LogResponse.__init__(self, headers, resp)


class DeleteConsumerGroupResponse(LogResponse):
    def __init__(self, headers, resp=''):
        LogResponse.__init__(self, headers, resp)


class ListConsumerGroupResponse(LogResponse):
    def __init__(self, resp, headers):
        LogResponse.__init__(self, headers, resp)
        self._count = len(resp)
        self.resp = resp
        self.consumer_groups = [ConsumerGroupEntity(group['name'], group['timeout'], group['order']) for group in
                                self.resp]

    def get_count(self):
        """

        :return:
        """
        return self._count

    @property
    def count(self):
        return self._count

    def get_consumer_groups(self):
        """

        :return:
        """
        return self.consumer_groups

    def log_print(self):
        """

        :return:
        """
        print('ListConsumerGroupResponse:')
        print('headers:', self.get_all_headers())
        print('count:', self._count)
        print('consumer_groups:', self.resp)


class UpdateConsumerGroupResponse(LogResponse):
    def __init__(self, headers, resp):
        LogResponse.__init__(self, headers, resp)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


class AclConfig(object):
    def __init__(self, aliyun_id, privilege_list=None):
        """ The acl config
        :type aliyun_id: string
        :param aliyun_id: the aliyun id e.g  "12312313131"

        :type privilege_list: string list
        :param privilege_list: the privilege list array, e.g ["WRITE", "READ", "ADMIN", "LIST"]
        """
        if privilege_list is None:
            privilege_list = []
        self.principle = aliyun_id
        self.privilege_list = privilege_list

    def to_json(self):
        json_value = {'principle': self.principle, 'privilege': self.privilege_list}
        return json_value

    def from_json(self, json_value):
        self.principle = json_value['principle']
        self.privilege_list = json_value['privilege']



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

__all__ = ['CreateShipperResponse', 'UpdateShipperResponse', 'DeleteShipperResponse',
           'GetShipperConfigResponse', 'ListShipperResponse', 'GetShipperTasksResponse',
           'RetryShipperTasksResponse']


from .logresponse import LogResponse
from .shipper_config import OdpsShipperConfig
from .shipper_config import OssShipperConfig
from .shipper_config import ShipperTask


class CreateShipperResponse(LogResponse):
    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('CreateShipperResponse:')
        print('headers:', self.get_all_headers())


class UpdateShipperResponse(LogResponse):
    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpdateShipperResponse:')
        print('headers:', self.get_all_headers())


class DeleteShipperResponse(LogResponse):
    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('DeleteShipperResponse:')
        print('headers:', self.get_all_headers())


class GetShipperConfigResponse(LogResponse):
    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.create_time = resp['createTime']
        self.last_modify_time = resp['lastModifyTime']
        self.type = resp['targetType']
        target_config = resp['targetConfiguration']
        if self.type == 'odps':
            self.config = OdpsShipperConfig(target_config["odpsEndpoint"],
                                            target_config["odpsProject"],
                                            target_config["odpsTable"],
                                            target_config["fields"],
                                            target_config["partitionColumn"],
                                            target_config["partitionTimeFormat"],
                                            target_config["bufferInterval"])
        elif self.type == 'oss':
            self.config = OssShipperConfig(target_config['ossBucket'],
                                           target_config['ossPrefix'],
                                           target_config['roleArn'],
                                           target_config['bufferInterval'],
                                           target_config['bufferSize'],
                                           target_config['compressType'])

    def get_config(self):
        """

        :return:
        """
        return self.config

    def get_create_time(self):
        """

        :return:
        """
        return self.create_time

    def get_last_modify_time(self):
        """

        :return:
        """
        return self.last_modify_time

    def log_print(self):
        """

        :return:
        """
        print('GetShipperConfigResponse:')
        print('type:' + self.type)
        print('config:' + str(self.config.to_json()))


class ListShipperResponse(LogResponse):
    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = resp['count']
        self.total = resp['total']
        self.shipper_names = resp['shipper']

    def get_shipper_count(self):
        """

        :return:
        """
        return self.count

    def get_shipper_total(self):
        """

        :return:
        """
        return self.total

    def get_shipper_names(self):
        """

        :return:
        """
        return self.shipper_names

    def log_print(self):
        """

        :return:
        """
        print('ListShipperResponse:')
        print('shipper count : ' + str(self.count))
        print('shipper total : ' + str(self.total))
        print('shipper names : ' + str(self.shipper_names))


class GetShipperTasksResponse(LogResponse):
    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = resp['count']
        self.total = resp['total']
        self.running_count = resp['statistics']['running']
        self.success_count = resp['statistics']['success']
        self.fail_count = resp['statistics']['fail']
        self.tasks = []
        for task_res in resp['tasks']:
            task = ShipperTask(task_res['id'], task_res['taskStatus'], task_res['taskMessage'],
                               task_res['taskCreateTime'],
                               task_res['taskLastDataReceiveTime'], task_res['taskFinishTime'])
            self.tasks.append(task)

    def get_task_count(self):
        """

        :return:
        """
        return self.count

    def get_count(self):
        return self.count

    def get_task_total(self):
        """

        :return:
        """
        return self.total

    def get_total(self):
        return self.total

    def get_running_task_count(self):
        """

        :return:
        """
        return self.running_count

    def get_success_task_count(self):
        """

        :return:
        """
        return self.success_count

    def get_fail_task_count(self):
        """

        :return:
        """
        return self.fail_count

    def _get_task_ids(self, status):
        task_ids = []
        for task in self.tasks:
            if task.task_status == status:
                task_ids.append(task.task_id)
        return task_ids

    def get_fail_task_ids(self):
        """

        :return:
        """
        return self._get_task_ids("fail")

    def get_running_task_ids(self):
        """

        :return:
        """
        return self._get_task_ids("running")

    def get_success_task_ids(self):
        """

        :return:
        """
        return self._get_task_ids("success")

    def get_tasks(self):
        """

        :return:
        """
        return self.tasks

    def log_print(self):
        """

        :return:
        """
        print('GetShipperTasksResponse:')
        print('ship count : ' + str(self.count))
        print('ship total : ' + str(self.total))
        print('ship running_count : ' + str(self.running_count))
        print('ship success_count : ' + str(self.success_count))
        print('ship fail_count : ' + str(self.fail_count))
        print('ship taks : ')
        for task in self.tasks:
            print(str(task.to_json()))

    def merge(self, response):
        if not isinstance(response, GetShipperTasksResponse):
            raise ValueError("passed response is not a GetShipperTasksResponse: " + str(type(response)))

        self.count += response.get_count()
        self.total = response.get_total()  # use the latest total count
        self.running_count += response.get_running_task_count()
        self.success_count += response.get_success_task_count()
        self.fail_count += response.get_fail_task_count()
        self.tasks.extend(response.get_tasks())

        # update body
        self.body['count'] = self.count
        self.body['total'] = self.total
        self.body['statistics']['running'] = self.running_count
        self.body['statistics']['success'] = self.success_count
        self.body['statistics']['fail'] = self.fail_count

        return self


class RetryShipperTasksResponse(LogResponse):
    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self,):
        print('RetryShipperTasksResponse')



from .trans_base import trans_comp_base
import csv
import six
import logging
import copy
import json
import jmespath
from jmespath.exceptions import ParseError
from ..exceptions import SettingError

__all__ = ['trans_comp_split_event']

logger = logging.getLogger(__name__)


def trans_comp_split_event(*args, **kwargs):
    if (args and isinstance(args[0], dict)) or ('event' in kwargs and isinstance(kwargs['event'], dict)):
        # event case
        split = split_event_transformer()
        return split(*args, **kwargs)
    else:
        return split_event_transformer(*args, **kwargs)


class split_event_transformer(trans_comp_base):
    DEFAULT_SEP = ','
    DEFAULT_QUOTE = '"'

    def __init__(self, sep=None, quote=None, lstrip=None, output=None, jmes=None):
        self.sep = sep or self.DEFAULT_SEP
        self.lstrip = True if lstrip is None else lstrip
        self.quote = quote or self.DEFAULT_QUOTE
        self.output_field = self._u(output)

        self.jmes = self._u(jmes or "")
        self.jmes_filter = None
        if jmes:
            try:
                self.jmes_filter = jmespath.compile(jmes)
            except jmespath.exceptions.ParseError as ex:
                raise SettingError(ex=ex, msg="Invalid JMES filter setting", settings=jmes)

    def _process_list(self, event, field, lst):
        if not lst:
            return event

        if len(lst) > 1:
            result = []
            for d in lst:
                e = copy.copy(event)
                e[field] = self._n(d)
                result.append(e)
            return result
        else:
            event[field] = self._n(lst[0])
            return event

    def _parse_list(self, v, flt=None):
        # process json first (via filter or directly list)
        try:
            ret = json.loads(v)

            # list and not jmes filter
            if isinstance(ret, (list,)) and not self.jmes:
                return ret

            # json object and jmes is configured
            if flt:
                ret = flt.search(ret)

                if isinstance(ret, (list,)):
                    # result is list as expected
                    return ret
                elif isinstance(ret, (six.text_type, six.binary_type)):
                    # result is just a string. then re-do the process and no jmes is passed this time
                    return self._parse_list(ret)
                else:
                    logger.info(
                        u'split_event_transformer: get unknown type of result with value "{0}" and jmes filter "{1}", skip it'.
                        format(v, self.jmes))
                    return None

            # else: go to next step: parse it as CSV

        except Exception as ex:
            # failed at json load or jmes filter. if jmes is configured, then it's an invalid event
            if flt:
                logger.info(u"split_event_transformer: value {0} is json or not invaid jmes settings {1}, skip it".
                            format(v, self.jmes))
                return None
            else:
                logger.debug(u"split_event_transformer: value {0} is not json, try to use csv".format(v))

        # continue to parse it as csv
        if isinstance(v, (six.text_type, six.binary_type)):
            result = list(csv.reader([v], skipinitialspace=self.lstrip, delimiter=self.sep, quotechar=self.quote))[0]
            return result
        else:
            logger.info(u"split_event_transformer: cannot extract list from value: {0}".format(v))
            return None

    def __call__(self, event, inpt):
        inpt = self._u(inpt)

        # overwrite input field
        if not self.output_field:
            self.output_field = inpt

        if inpt in event:
            # json/csv mode
            ret = self._parse_list(event[inpt], self.jmes_filter)
            return self._process_list(event, self.output_field, ret)
        else:
            logger.error(u"trans_comp_lookup: unknown type of input field {0}".format(inpt))

        return event



from .trans_base import trans_comp_check_mdoe_base
import six
import os
from ..exceptions import SettingError
import csv
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
import logging
import inspect

logger = logging.getLogger(__name__)

__all__ = ['trans_comp_lookup']


class ETLLookupError(SettingError):
    pass


class Table(object):
    def __init__(self, reader, case_insensitive):
        self.field_names = set(reader.fieldnames)
        self.rows = [row for row in reader]
        self.case_insensitive = case_insensitive

    def get_row(self, inputs):
        assert isinstance(inputs, dict), ETLLookupError(msg="trans_comp_lookup: inputs are not dict as expected",
                                                        settings=inputs)

        for k, v in six.iteritems(inputs):
            if k not in self.field_names:
                logger.info(u"trans_comp_lookup: key {0} doesn't exist in existing fields names: {1}".format(k,
                                                                                                            self.field_names))
                return None

        for row in self.rows:
            for k, v in six.iteritems(inputs):
                if row[k] != '*' and (row[k] != v if not self.case_insensitive else row[k].lower() != v.lower()):
                    break
            else:
                return row

        return None


class DefaultDict(object):
    def __init__(self, dct, case_insensitive):
        self.default = dct.get('*', None)
        self.case_insensitive = case_insensitive
        if case_insensitive:
            self.data = dict((k.lower(), v) for k, v in six.iteritems(dct))
        else:
            self.data = dct

    def __getitem__(self, item):
        item = item.lower() if self.case_insensitive else item
        return self.data.get(item, self.default)

    def __contains__(self, item):
        item = item.lower() if self.case_insensitive else item
        return item in self.data or self.default is not None


class trans_comp_lookup(trans_comp_check_mdoe_base):
    EXTERNAL_CACHE = {}

    def __init__(self, data, output_fields, sep=',', quote='"', lstrip=True, case_insensitive=True, headers=None, mode=None):
        super(trans_comp_lookup, self).__init__(mode=mode)

        data = self._u(data)
        output_fields = self._u(output_fields)
        sep = sep
        quote = quote
        headers = self._u(headers)

        if isinstance(data, dict):
            self.data = DefaultDict(data, case_insensitive)

            # init output fields
            if isinstance(output_fields, (six.binary_type, six.text_type)):
                self.output_fields = [output_fields]
            elif isinstance(output_fields, Iterable):
                self.output_fields = []
                for x in output_fields:
                    if isinstance(output_fields, (six.binary_type, six.text_type)):
                        self.output_fields.append(x)
                    else:
                        raise SettingError(settings=output_fields)
            else:
                raise SettingError(settings=output_fields)

        elif isinstance(data, (six.binary_type, six.text_type)):

            # parse headers
            if headers:
                if isinstance(headers, (six.text_type, six.binary_type)):
                    headers = tuple(x.strip() for x in headers.split(","))
                else:
                    headers = tuple(headers)

            self.sig = (data, sep, quote, lstrip, headers)

            if self.sig in self.EXTERNAL_CACHE:
                self.data = self.EXTERNAL_CACHE[self.sig]
            else:
                type_path = data.split("://")
                file_type = type_path[0] if len(type_path) == 2 else u'file'
                file_path = type_path[1] if len(type_path) == 2 else data
                if file_type != 'file':
                    raise SettingError(msg="trans_comp_lookup: unsupported file type", settings=data)
                if not os.path.isfile(file_path):
                    caller_frame_record = inspect.stack()[1]
                    module = inspect.getmodule(caller_frame_record[0])
                    file_path = os.path.sep.join([os.path.dirname(module.__file__), file_path])
                    if not os.path.isfile(file_path):
                        raise SettingError(msg="trans_comp_lookup: cannot locate the file path", settings=data)

                with open(file_path) as file:
                    reader = csv.DictReader(file, fieldnames=headers, skipinitialspace=lstrip, delimiter=sep, quotechar=quote)
                    self.data = Table(reader, case_insensitive)

                # put into cache for re-use for other calling
                self.EXTERNAL_CACHE[self.sig] = self.data

            # init output fields
            self.output_fields = {}
            if isinstance(output_fields, (six.binary_type, six.text_type)):
                self.output_fields[output_fields] = output_fields
            elif isinstance(output_fields, Iterable):
                for f in output_fields:
                    if isinstance(f, (six.binary_type, six.text_type)):
                        self.output_fields[f] = f
                    elif isinstance(f, (tuple, list)) and len(f) == 2:
                        self.output_fields[f[0]] = f[1]
                    else:
                        raise SettingError(settings=output_fields)

                for f in self.output_fields:
                    assert f in self.data.field_names, SettingError(
                        msg="trans_comp_lookup: output field {0} doesn't exist in lookup table".format(f))
            else:
                raise SettingError(settings=output_fields)
        else:
            raise SettingError(settings=data)

    def __call__(self, event, inpt):
        inpt = self._u(inpt)

        if isinstance(self.data, DefaultDict):
            # simple dict mode
            if isinstance(inpt, (six.binary_type, six.text_type)):
                inpt = [inpt]

            if isinstance(inpt, Iterable):
                for i in inpt:
                    if not isinstance(i, (six.binary_type, six.text_type)):
                        logger.error(u'trans_comp_lookup: type of input field "{0}" is unknown'.format(i))
                        continue

                    if i not in event:
                        logger.info(u'trans_comp_lookup: event "{0}" doesn not contain field "{1}"'.format(event, i))
                        continue

                    # get input value
                    f_v = event[i]

                    # set output value
                    for f_n in self.output_fields:
                        if f_v in self.data:
                            self.set(event, f_n, self.data[f_v])
                        else:
                            logger.info(u'trans_comp_lookup: value {0} not exit in lookup {1}'.format(f_v, self.data))
            else:
                logger.error(u"trans_comp_lookup: unknown type of input field {0}".format(inpt))
        else:
            # lookup type
            if isinstance(inpt, (six.binary_type, six.text_type)):
                inpt = [inpt]

            if isinstance(inpt, Iterable):
                inpt_map = {}
                for i in inpt:
                    if not isinstance(i, (six.binary_type, six.text_type, list, tuple)):
                        logger.error(u'trans_comp_lookup: type of input field "{0}" is unknown'.format(i))
                        # must exit, or else skip it for lookup type
                        return event

                    if isinstance(i, (tuple, list)) and len(i) != 2:
                        logger.error(u'trans_comp_lookup: type of input field "{0}" is unsupported'.format(i))
                        # must exit, or else skip it for lookup type
                        return event

                    if isinstance(i, (six.binary_type, six.text_type)):
                        i = (i, i)

                    if i[0] not in event:
                        logger.info(u'trans_comp_lookup: event "{0}" doesn not contain field "{1}"'.format(event, i[0]))
                        # must exit, or else skip it for lookup type
                        return event

                    inpt_map[i[1]] = event[i[0]]

                row = self.data.get_row(inpt_map)
                if row is None:
                    logger.info(
                        u'trans_comp_lookup: cannot find proper value for inpt "{0}" in event "{0}" doesn not contain field "{1}"'.format(
                            inpt_map, event))
                    return event

                for f, f_new in six.iteritems(self.output_fields):
                    if f in row:
                        self.set(event, f_new, row[f])
                    else:
                        logger.info(u"trans_comp_lookup: field {0} doesn't exit in lookup row {1}".format(f, row))

            else:
                logger.error(u"trans_comp_lookup: unknown type of input field {0}".format(inpt))

        return event



import six

from .transform_base import transform_base
from ..etl_util import get_re_full_match
from ..exceptions import SettingError
from ..trans_comp import KV
from ..etl_util import support_event_list_simple, u

__all__ = ["simple_drop", "simple_keep", 'drop_fields', "keep_fields", "rename_fields", 'extract_kv_fields',
           'DROP', 'KEEP', 'ALIAS', 'RENAME', 'DROP_F', 'KEEP_F', 'KV_F']


class simple_drop(transform_base):
    @support_event_list_simple
    def __call__(self, event):
        return None


class simple_keep(transform_base):
    @support_event_list_simple
    def __call__(self, event):
        return event


class drop_fields(transform_base):
    def __init__(self, config):
        if isinstance(config, (six.text_type, six.binary_type)):
            self.check = get_re_full_match(config)
        elif isinstance(config, list):  # string list
            checks = [get_re_full_match(c) for c in config]
            self.check = lambda k: any(ck(k) for ck in checks)
        else:
            raise SettingError(None, "keep_fields setting {0} is not supported".format(config))

    @support_event_list_simple
    def __call__(self, event):
        return dict((k, v) for k, v in six.iteritems(event) if not self.check(k))


class keep_fields(transform_base):
    def __init__(self, config):
        config = u(config)

        if isinstance(config, (six.text_type, six.binary_type)):
            self.check = get_re_full_match(config)
        elif isinstance(config, list):  # string list
            checks = [get_re_full_match(c) for c in config]
            self.check = lambda k: any(ck(k) for ck in checks)
        else:
            raise SettingError(None, "keep_fields setting {0} is not supported".format(config))

    @support_event_list_simple
    def __call__(self, event):
        return dict((k, v) for k, v in six.iteritems(event) if self.check(k))


class rename_fields(transform_base):
    def __init__(self, config):
        config = u(config)

        if isinstance(config, (dict, )):
            try:
                config_ptn = [(get_re_full_match(k), v) for k, v in six.iteritems(config)]

                def check(k):
                    for c, new_k in config_ptn:
                        if c(k):
                            return new_k
                    return k

                self.new_name = check
            except Exception as ex:
                raise SettingError(ex, "rename setting {0} is not invalid".format(config))
        elif config is None or config == "":
            self.new_name = lambda k: k
        else:
            raise SettingError(None, "rename setting {0} is not supported, should be dict type".format(config))

    @support_event_list_simple
    def __call__(self, event):
        return dict((self.new_name(k), v) for k, v in six.iteritems(event))


class extract_kv_fields(transform_base):
    def __init__(self, config):
        config = u(config)

        if isinstance(config, (six.text_type, six.binary_type)):
            self.check = get_re_full_match(config)
        elif isinstance(config, list):  # string list
            checks = [get_re_full_match(c) for c in config]
            self.check = lambda k: any(ck(k) for ck in checks)
        else:
            raise SettingError(None, "extract_kv setting {0} is not supported".format(config))

    @support_event_list_simple
    def __call__(self, event):
        fields_list = [k for k, v in six.iteritems(event) if self.check(k)]
        return KV(event, fields_list)


DROP = simple_drop()
KEEP = simple_keep()

ALIAS, RENAME = rename_fields, rename_fields
DROP_F = drop_fields
KEEP_F = keep_fields
KV_F = extract_kv_fields



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

import time

from .util import Util


class MachineGroupDetail(object):
    """ The machine group detail info

    :type group_name: string
    :param group_name: group name

    :type machine_type: string
    :param machine_type: "ip" or "userdefined"

    :type machine_list: string list
    :param machine_list: the list of machine ips or machine userdefined, e.g ["127.0.0.1", "127.0.0.2"]

    :type group_type: string
    :param group_type: the machine group type, "" or "Armory"

    :type group_attribute: dict
    :param group_attribute: the attributes in group, it contains two optional key :
            1. "externalName": only used if the group_type is "Armory", its the Armory name
            2. "groupTopic": group topic value
    """

    def __init__(self, group_name=None, machine_type=None, machine_list=None, group_type="", group_attribute=None):
        if group_attribute is None:
            group_attribute = {}
        self.group_name = group_name
        self.machine_type = machine_type
        self.machine_list = machine_list
        self.group_type = group_type
        self.group_attribute = group_attribute
        self.create_time = int(time.time())
        self.last_modify_time = self.create_time

    def to_json(self):
        json_value = {'groupName': self.group_name, 'groupType': self.group_type,
                      'groupAttribute': self.group_attribute, 'machineIdentifyType': self.machine_type,
                      'machineList': self.machine_list}
        return json_value

    def from_json(self, json_value):
        self.group_name = json_value.get("groupName", json_value.get("group_name", None))
        self.group_type = json_value.get("groupType", json_value.get("group_type", ""))
        self.machine_type = json_value.get("machineIdentifyType", json_value.get("machine_type", None))
        self.group_attribute = json_value.get("groupAttribute", json_value.get("group_attribute", {}))
        self.machine_list = json_value.get("machineList", json_value.get("machine_list", None))
        self.create_time = json_value.get("crateTime", None)
        self.last_modify_time = json_value.get("lastModifyTime", None)


class MachineStatus(object):
    """ the machine status
    :type ip: string
    :param ip: the machine ip

    :type machine_unique_id: string
    :param machine_unique_id: the machine unique id
    
    :type user_defined_id: string
    :param user_defined_id: the user defined id

    :type heartbeat_time: int
    :param heartbeat_time: last updated heartbeat_time
    """

    def __init__(self, ip, machine_unique_id, user_defined_id, heartbeat_time):
        self.ip = ip
        self.machine_unique_id = machine_unique_id
        self.user_defined_id = user_defined_id
        self.heartbeat_time = heartbeat_time



import logging
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import re
import six

from .trans_base import trans_comp_check_mdoe_base
from ..etl_util import cached, get_re_full_match, get_set_mode_if_skip_fn

__all__ = ['trans_comp_kv']

logger = logging.getLogger(__name__)


def trans_comp_kv(*args, **kwargs):
    if (args and isinstance(args[0], dict)) or ('event' in kwargs and isinstance(kwargs['event'], dict)):
        # event case
        kv = kv_transformer()
        return kv(*args, **kwargs)
    else:
        return kv_transformer(*args, **kwargs)


class kv_transformer(trans_comp_check_mdoe_base):
    DEFAULT_SEP = u'='
    DEFAULT_QUOTE = u'"'

    @staticmethod
    @cached
    def _get_kv_ptn(sep, quote, escape):
        p1 = u'(?!{0})([\u4e00-\u9fa5\u0800-\u4e00\\w\\.\\-]+)\\s*{0}\\s*([\u4e00-\u9fa5\u0800-\u4e00\\w\\.\\-]+)'
        if not escape:
            p2 = u'(?!{0})([\u4e00-\u9fa5\u0800-\u4e00\\w\\.\\-]+)\\s*{0}\\s*{1}\s*([^{1}]*?)\s*{1}'
        else:
            p2 = u'(?!{0})([\u4e00-\u9fa5\u0800-\u4e00\\w\\.\\-]+)\\s*{0}\\s*{1}\s*((?:[^{1}]|\\\\{1})*?[^\\\\]){1}'
        ps = u'|'.join([p1, p2]).format(sep, quote)

        logger.info(u"trans_comp_kv: get ptn: {0}".format(ps))
        return re.compile(ps)

    def __init__(self, prefix=None, suffix=None, sep=None, quote=None, escape=None, mode=None):
        super(kv_transformer, self).__init__(mode=mode)
        self.prefix = self._u("" if prefix is None else prefix)
        self.suffix = self._u("" if suffix is None else suffix)
        self.escape = False if escape is None else escape

        self.sep = self._u(self.DEFAULT_SEP if sep is None else sep)
        self.quote = self._u(self.DEFAULT_QUOTE if quote is None else quote)
        self.ptn = self._get_kv_ptn(self.sep, self.quote, self.escape)
        self.normalize = None

    def set(self, e, k, v, real_k=None, check_kw_name=False):
        """override base to handle escape case: replace \" to " """
        if self.escape:
            v = v.strip().replace("\\" + self.quote, self.quote)
        return super(kv_transformer, self).set(e, k, v, real_k=real_k, check_kw_name=check_kw_name)

    def _extract_kv(self, event, value):
        if isinstance(value, six.binary_type):
            value = value.decode('utf8', 'ignore')

        r = self.ptn.findall(value)

        new_event = {}
        for k1, v1, k2, v2 in r:
            if k1:
                self.set(event, k1, v1, real_k=u"{0}{1}{2}".format(self.prefix, k1, self.suffix), check_kw_name=True)
            if k2:
                self.set(event, k2, v2, real_k=u"{0}{1}{2}".format(self.prefix, k2, self.suffix), check_kw_name=True)

        return new_event

    def __call__(self, event, inpt):
        inpt = self._u(inpt)

        # simple dict mode
        if isinstance(inpt, (six.binary_type, six.text_type)):
            inpt = [inpt]

        if isinstance(inpt, Iterable):
            for i in inpt:
                if not isinstance(i, (six.binary_type, six.text_type)):
                    logger.error(u'trans_comp_lookup: type of input field "{0}" is unknown'.format(i))
                    continue

                if i not in event:
                    logger.info(u'trans_comp_lookup: event "{0}" doesn not contain field "{1}"'.format(event, i))
                    continue

                # get input value
                new_event = self._extract_kv(event, event[i])
                event.update(new_event)
        else:
            logger.error(u"trans_comp_lookup: unknown type of input field {0}".format(inpt))

        return event



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

"""Setup script for log service SDK.

You need to install google protocol buffer, setuptools and python-requests.
https://code.google.com/p/protobuf/
https://pypi.python.org/pypi/setuptools
http://docs.python-requests.org/

Depending on your version of Python, these libraries may also should be installed:
http://pypi.python.org/pypi/simplejson/

"""

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


import sys
import re

install_requires_common = ['requests', 'six', 'python-dateutil', 'elasticsearch', 'jmespath']
install_requires = []

if sys.version_info[:2] == (2, 6):
    install_requires = ['protobuf <= 3.4.0', 'enum34', 'futures']
elif sys.version_info[0] == 2:
    install_requires = ['protobuf <= 3.17.3', 'enum34', 'futures', 'dateparser <= 0.7.6']
elif sys.version_info[:2] == (3, 3):
    install_requires = ['protobuf>=3.4.0', 'enum34',  'dateparser']
elif sys.version_info[0] == 3:
    install_requires = ['protobuf>=3.4.0',  'dateparser']

install_requires.extend(install_requires_common)

packages = [
            'aliyun',
            'aliyun.log',
            'aliyun.log.ext',
            'aliyun.log.etl_core',
            'aliyun.log.etl_core.transform',
            'aliyun.log.etl_core.trans_comp',
            'aliyun.log.consumer',
            'aliyun.log.es_migration'
            ]

version = ''
with open('aliyun/log/version.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

classifiers = [
            'Development Status :: 5 - Production/Stable',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: Implementation :: PyPy'
    ]


long_description = """
Python SDK for Alicloud Log Service 
http://aliyun-log-python-sdk.readthedocs.io
"""

setup(
      name='aliyun-log-python-sdk',
      version=version,
      description='Aliyun log service Python client SDK',
      author='Aliyun',
      url='https://github.com/aliyun/aliyun-log-python-sdk',
      install_requires=install_requires,
      packages=packages,
      classifiers=classifiers,
      long_description=long_description
     )



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


class Histogram(object):
    """
    The class used to present the result of log histogram status. For every log
    histogram, it contains : from/to time range, hit log count and query
    completed status.
    
    :type fromTime: int
    :param fromTime: the begin time
    
    :type toTime: int
    :param toTime: the end time
    
    :type count: int
    :param count: log count of histogram that query hits 
    
    :type progress: string
    :param progress: histogram query status(Complete or InComplete)
    """

    def __init__(self, fromTime, toTime, count, progress):
        self.fromTime = fromTime
        self.toTime = toTime
        self.count = count
        self.progress = progress

    def get_from(self):
        """ Get begin time
        
        :return: int, begin time
        """
        return self.fromTime

    def get_to(self):
        """ Get end time
        
        :return: int, end time
        """
        return self.toTime

    def get_count(self):
        """ Get log count of histogram that query hits
        
        :return: int, log count of histogram that query hits
        """
        return self.count

    def is_completed(self):
        """ Check if the histogram is completed
        
        :return: bool, true if this histogram is completed
        """
        return self.progress == 'Complete'

    def log_print(self):
        print('Histogram:')
        print('from:', self.fromTime)
        print('to:', self.toTime)
        print('count:', self.count)
        print('progress:', self.progress)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logrequest import LogRequest


class PutLogsRequest(LogRequest):
    """ The request used to send data to log. 
    
    :type project: string
    :param project: project name
    
    :type logstore: string
    :param logstore: logstore name
    
    :type topic: string
    :param topic: topic name

    :type source: string
    :param source: source of the logs
    
    :type logitems: list<LogItem>
    :param logitems: log data

    :type hashKey: String
    :param hashKey: put data with set hash, the data will be send to shard whose range contains the hashKey

    :type compress: bool
    :param compress: if need to compress the logs

    :type logtags: list
    :param logtags: list of key:value tag pair , [(tag_key_1,tag_value_1) , (tag_key_2,tag_value_2)]

    """

    def __init__(self, project=None, logstore=None, topic=None, source=None, logitems=None, hashKey=None,
                 compress=True, logtags=None):
        LogRequest.__init__(self, project)
        self.logstore = logstore
        self.topic = topic
        self.source = source
        self.logitems = logitems
        self.hashkey = hashKey
        self.compress = compress
        self.logtags = logtags

    def get_compress(self):
        return self.compress

    def set_compress(self, compress):
        self.compress = compress

    def get_logstore(self):
        """ Get logstore name
        
        :return: string, logstore name
        """
        return self.logstore if self.logstore else ''

    def set_logstore(self, logstore):
        """ Set logstore name
        
        :type logstore: string
        :param logstore: logstore name
        """
        self.logstore = logstore

    def get_topic(self):
        """ Get topic name
        
        :return: string, topic name
        """
        return self.topic if self.topic else ''

    def set_topic(self, topic):
        """ Set topic name
        
        :type topic: string
        :param topic: topic name
        """
        self.topic = topic

    def get_source(self):
        """ Get log source
        
        :return: string, log source
        """
        return self.source

    def set_source(self, source):
        """ Set log source
        
        :type source: string
        :param source: log source
        """
        self.source = source

    def get_log_items(self):
        """ Get all the log data
        
        :return: LogItem list, log data
        """
        return self.logitems

    def set_log_items(self, logitems):
        """ Set the log data
        
        :type logitems: LogItem list
        :param logitems: log data
        """
        self.logitems = logitems

    def get_log_tags(self):
        """ Get all the log tags
        
        :return: Logtags list, log data
        """
        return self.logtags

    def set_log_tags(self, logtags):
        """ Set the log tags
        
        :type logtags: logtags list
        :param logtags: log tags
        """
        self.logtags = logtags

    def set_hash_key(self, hashKey):
        self.hashkey = hashKey

    def get_hash_key(self):
        return self.hashkey



import json
from aliyun.log.util import Util


def check_type_for_init(**kwargs):
    type_dict = kwargs["instance"].type_dict
    del kwargs["instance"]
    for key, value in kwargs.items():
        if value is not None:
            need_type = type_dict.get(key)
            if not isinstance(value, need_type):
                raise TypeError("the %s type must be %s" % (key, need_type))


def check_params(name, need_type):
    def outer(func):
        def inner(*args, **kwargs):
            value = None
            if len(args) > 1:
                value = args[1]
            if kwargs:
                value = kwargs.get(name)
            if value is not None:
                if not isinstance(value, need_type):
                    raise TypeError("the %s type must be %s" % (name, need_type))
            func(*args, **kwargs)

        return inner

    return outer


class ResourceRecord:
    """
     ResourceRecord
     Create required:value
     Update required:id,value

    :type value: dict
    :param value: record value

    :type record_id: string
    :param record_id: record id

    :type tag: string
    :param tag: record tag
     """

    type_dict = {
        "record_id": str,
        "tag": str,
        "value": dict
    }

    def __init__(self, record_id=None, tag=None, value=None):
        check_type_for_init(record_id=record_id, tag=tag, value=value, instance=self)
        self.record_id = record_id
        self.tag = tag
        self.value = value
        self.create_time = None
        self.last_modify_time = None

    def get_record_id(self):
        return self.record_id

    @check_params("record_id", str)
    def set_record_id(self, record_id):
        self.record_id = record_id

    def get_tag(self):
        return self.tag

    @check_params("tag", str)
    def set_tag(self, tag):
        self.tag = tag

    def get_value(self):
        return self.value

    @check_params("tag", dict)
    def set_value(self, value):
        self.value = value

    def get_create_time(self):
        return self.create_time

    def set_create_time(self, create_time):
        self.create_time = create_time

    def get_last_modify_time(self):
        return self.last_modify_time

    def set_last_modify_time(self, last_modify_time):
        self.last_modify_time = last_modify_time

    def to_dict(self):
        result = {"value": self.value}
        if self.tag:
            result["tag"] = self.tag
        if self.record_id:
            result["id"] = self.record_id
        if self.create_time:
            result["create_time"] = self.create_time
        if self.last_modify_time:
            result["last_modify_time"] = self.last_modify_time
        return result

    def check_value(self):
        if not self.value:
            raise ValueError("ResourceRecord Value must not be None")

    def check_for_update(self):
        self.check_value()
        if not self.record_id:
            raise ValueError(" ResourceRecord id must not be None")

    @classmethod
    def from_dict(cls, dict_data):
        return ResourceRecord(record_id=dict_data.get("id"), tag=dict_data.get("tag"),
                              value=json.loads(dict_data.get("value")))


class ResourceSchemaItem:
    """
    ResourceSchemaItem
   :type column: string
   :param column: column

   :type ext_info: string
   :param ext_info: schema ext_info

   :type schema_type: string
   :param schema_type: schema_type
    """
    type_dict = {
        "column": str,
        "ext_info": str,
        "schema_type": str
    }

    def __init__(self, column=None, ext_info=None, schema_type=None):
        check_type_for_init(column=column, ext_info=ext_info, schema_type=schema_type, instance=self)
        self.column = column
        self.ext_info = ext_info
        self.type = schema_type

    def get_column(self):
        return self.column

    @check_params("column", str)
    def set_column(self, column):
        self.column = column

    def get_ext_info(self):
        return self.ext_info

    @check_params("ext_info", str)
    def set_ext_info(self, ext_info):
        self.ext_info = ext_info

    def get_schema_type(self):
        return self.type

    @check_params("schema_type", str)
    def set_schema_type(self, schema_type):
        self.type = schema_type

    def to_dict(self):
        schema_dict = {}
        if self.get_column():
            schema_dict["column"] = self.get_column()
        if self.get_ext_info():
            schema_dict["ext_info"] = self.get_column()
        if self.get_schema_type():
            schema_dict["type"] = self.get_schema_type()
        return schema_dict


class Resource:
    """ Resource
    Create required: resource_name resource_type
    Update required: resource_name
    :type resource_name: string
    :param resource_name: resource name

    :type resource_type: string
    :param resource_type: resource type

    :type description: string
    :param description: the description of resource name

    :type schema_list: List[ResourceSchemaItem]
    :param schema_list: schema of resource

    :type acl: dict
    :param acl: policy  example:{"policy": {"type": "all_rw"}}

    :type ext_info: string
    :param ext_info: extra info of this resource
    """

    type_dict = {
        "resource_name": str,
        "resource_type": str,
        "schema_list": list,
        "description": str,
        "acl": dict,
        "ext_info": str
    }

    def __init__(self, resource_name=None, resource_type=None, schema_list=None,
                 description=None, acl=None, ext_info=None):
        check_type_for_init(resource_name=resource_name, resource_type=resource_type, schema_list=schema_list,
                            description=description, acl=acl, ext_info=ext_info, instance=self)
        self.resource_name = resource_name
        self.resource_type = resource_type
        self.description = description
        self.schema_list = self._check_schema(schema_list) if schema_list else schema_list
        self.acl = acl
        self.ext_info = ext_info
        self.create_time = None
        self.last_modify_time = None

    def get_resource_name(self):
        return self.resource_name

    @check_params("resource_name", str)
    def set_resource_name(self, resource_name):
        self.resource_name = resource_name

    def get_resource_type(self):
        return self.resource_type

    @check_params("resource_type", str)
    def set_resource_type(self, resource_type):
        self.resource_type = resource_type

    def get_description(self):
        return self.description

    @check_params("description", str)
    def set_description(self, description):
        self.description = description

    def get_schema(self):
        return self.schema_list

    @check_params("description", list)
    def set_schema_list(self, schema):
        self.schema_list = self._check_schema(schema) if schema else schema

    def get_acl(self):
        return self.acl

    @check_params("acl", dict)
    def set_acl(self, acl):
        self.acl = acl

    def get_ext_info(self):
        return self.ext_info

    @check_params("ext_info", str)
    def set_ext_info(self, ext_info):
        self.ext_info = ext_info

    def get_create_time(self):
        return self.create_time

    def set_create_time(self, create_time):
        self.create_time = create_time

    def get_last_modify_time(self):
        return self.last_modify_time

    def set_last_modify_time(self, last_modify_time):
        self.last_modify_time = last_modify_time

    def check_for_update(self):
        if not self.resource_name:
            raise ValueError("resource_name must not be None!")

    def check_for_create(self):
        if not (self.resource_name and self.resource_type):
            raise ValueError("resource_name and resource_type must not be None!")

    @staticmethod
    def _check_schema(schema_list):
        for schema in schema_list:
            if not isinstance(schema, ResourceSchemaItem):
                raise TypeError("schema list element must be instance of ResourceSchemaItem ")
        return schema_list

    def to_dict(self):
        schema_list = [schema.to_dict() for schema in self.schema_list]
        result = {"schema": schema_list}
        if self.create_time:
            result["create_time"] = self.create_time
        if self.last_modify_time:
            result["last_modify_time"] = self.last_modify_time
        if self.last_modify_time:
            result["last_modify_time"] = self.last_modify_time
        if self.resource_name:
            result["resource_name"] = self.resource_name
        if self.resource_type:
            result["resource_type"] = self.resource_type
        if self.description:
            result["description"] = self.description
        if self.ext_info:
            result["ext_info"] = self.ext_info
        if self.acl:
            result["acl"] = self.acl
        return result

    @classmethod
    def from_dict(cls, dict_data):
        schema_list = Util.convert_unicode_to_str(json.loads(dict_data.get("schema"))).get("schema")
        schema_instance_list = []
        for schema in schema_list:
            schema_instance_list.append(ResourceSchemaItem(column=schema.get("column"), schema_type=schema.get("type"),
                                                           ext_info=schema.get("ext_info")))
        resource = Resource()
        resource.set_resource_name(dict_data.get("name"))
        resource.set_description(dict_data.get('description'))
        resource.set_resource_type(dict_data.get("type"))
        resource.set_ext_info(dict_data.get("extInfo"))
        resource.set_acl(json.loads(dict_data.get("acl")))
        resource.set_schema_list(schema_instance_list)
        resource.set_create_time(dict_data.get("create_time"))
        resource.set_last_modify_time(dict_data.get("last_modify_time"))
        return resource



from .trans_regex import *
from .trans_lookup import *
from .trans_csv import *
from .trans_kv import *
from .trans_json import *
from .trans_mv import *
from .trans_set_field import *

__all__ = ['LOOKUP', 'REGEX', 'CSV', 'TSV', 'PSV', 'JSON', 'KV', 'V', 'SPLIT', 'ZIP']

# field based
LOOKUP = trans_comp_lookup
REGEX = trans_comp_regex
CSV = trans_comp_csv
TSV = trans_comp_tsv
PSV = trans_comp_psv
JSON = trans_comp_json
KV = trans_comp_kv
SPLIT = trans_comp_split_event
ZIP = trans_set_field_zip


#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logrequest import LogRequest


class GetLogsRequest(LogRequest):
    """ The request used to get logs by a query from log.
    
    :type project: string
    :param project: project name
    
    :type logstore: string
    :param logstore: logstore name
    
    :type fromTime: int/string
    :param fromTime: the begin time, or format of time in format "%Y-%m-%d %H:%M:%S" e.g. "2018-01-02 12:12:10"
    
    :type toTime: int/string
    :param toTime: the end time, or format of time in format "%Y-%m-%d %H:%M:%S" e.g. "2018-01-02 12:12:10"

    :type topic: string
    :param topic: topic name of logs
    
    :type query: string
    :param query: user defined query
    
    :type line: int
    :param line: max line number of return logs
    
    :type offset: int
    :param offset: line offset of return logs
    
    :type reverse: bool
    :param reverse: if reverse is set to true, the query will return the latest logs first

    :type power_sql: bool
    :param power_sql: if power_sql is set to true, the query will run on enhanced sql mode
    """

    def __init__(self, project=None, logstore=None, fromTime=None, toTime=None, topic=None,
                 query=None, line=100, offset=0, reverse=False, power_sql=False):
        LogRequest.__init__(self, project)
        self.logstore = logstore
        self.fromTime = fromTime
        self.toTime = toTime
        self.topic = topic
        self.query = query
        self.line = line
        self.offset = offset
        self.reverse = reverse
        self.power_sql = power_sql

    def get_logstore(self):
        """ Get logstore name
        
        :return: string, logstore name
        """
        return self.logstore if self.logstore else ''

    def set_logstore(self, logstore):
        """ Set logstore name
        
        :type logstore: string
        :param logstore: logstore name
        """
        self.logstore = logstore

    def get_topic(self):
        """ Get topic name
        
        :return: string, topic name
        """
        return self.topic if self.topic else ''

    def set_topic(self, topic):
        """ Set topic name
        
        :type topic: string
        :param topic: topic name
        """
        self.topic = topic

    def get_from(self):
        """ Get begin time
        
        :return: int, begin time
        """
        return self.fromTime

    def set_from(self, fromTime):
        """ Set begin time
        
        :type fromTime: int
        :param fromTime: begin time
        """
        self.fromTime = fromTime

    def get_to(self):
        """ Get end time
        
        :return: int, end time
        """
        return self.toTime

    def set_to(self, toTime):
        """ Set end time
        
        :type toTime: int
        :param toTime: end time
        """
        self.toTime = toTime

    def get_query(self):
        """ Get user defined query
        
        :return: string, user defined query
        """
        return self.query

    def set_query(self, query):
        """ Set user defined query
        
        :type query: string
        :param query: user defined query
        """
        self.query = query

    def get_line(self):
        """ Get max line number of return logs
        
        :return: int, max line number of return logs
        """
        return self.line

    def set_line(self, line):
        """ Set max line number of return logs
        
        :type line: int
        :param line: max line number of return logs
        """
        self.line = line

    def get_offset(self):
        """ Get line offset of return logs
        
        :return: int, line offset of return logs
        """
        return self.offset

    def set_offset(self, offset):
        """ Set line offset of return logs
        
        :type offset: int
        :param offset: line offset of return logs
        """
        self.offset = offset

    def get_reverse(self):
        """ Get request reverse flag
        
        :return: bool, reverse flag
        """
        return self.reverse

    def set_reverse(self, reverse):
        """ Set request reverse flag
        
        :type reverse: bool
        :param reverse: reverse flag
        """
        self.reverse = reverse

    def get_power_sql(self):
        """ Get request power_sql flag

        :return: bool, power_sql flag
        """
        return self.power_sql

    def set_power_sql(self, power_sql):
        """ Set request power_sql flag

        :type power_sql: bool
        :param power_sql: power_sql flag
        """
        self.power_sql = power_sql


class GetProjectLogsRequest(LogRequest):
    """ The request used to get logs by a query from log cross multiple logstores.

    :type project: string
    :param project: project name

    :type query: string
    :param query: user defined query

    :type power_sql: bool
    :param power_sql: if power_sql is set to true, the query will run on enhanced sql mode
    """

    def __init__(self, project=None, query=None, power_sql=False):
        LogRequest.__init__(self, project)
        self.query = query
        self.power_sql = power_sql

    def get_query(self):
        """ Get user defined query

        :return: string, user defined query
        """
        return self.query

    def set_query(self, query):
        """ Set user defined query

        :type query: string
        :param query: user defined query
        """
        self.query = query

    def get_power_sql(self):
        """ Get request power_sql flag

        :return: bool, power_sql flag
        """
        return self.power_sql

    def set_power_sql(self, power_sql):
        """ Set request power_sql flag

        :type power_sql: bool
        :param power_sql: power_sql flag
        """
        self.power_sql = power_sql
